# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Shard downloading from various storage providers."""

import os
import pathlib
import shutil
import urllib.parse
from time import sleep, time
from typing import Any, Optional

from streaming.base.util import get_import_exception_message

__all__ = [
    'download_from_s3',
    'download_from_sftp',
    'download_from_gcs',
    'download_from_oci',
    'download_from_azure',
    'download_from_azure_datalake',
    'download_from_hf',
    'download_from_databricks_unity_catalog',
    'download_from_dbfs',
    'download_from_alipan',
    'download_from_local',
]

BOTOCORE_CLIENT_ERROR_CODES = {'403', '404', 'NoSuchKey'}

GCS_ERROR_NO_AUTHENTICATION = """\
Either set the environment variables `GCS_KEY` and `GCS_SECRET` or use any of the methods in \
https://cloud.google.com/docs/authentication/external/set-up-adc to set up Application Default \
Credentials. See also https://docs.mosaicml.com/projects/mcli/en/latest/resources/secrets/gcp.html.
"""


def download_from_s3(remote: str, local: str, timeout: float) -> None:
    """Download a file from remote AWS S3 (or any S3 compatible object store) to local.

    Args:
        remote (str): Remote path (S3).
        local (str): Local path (local filesystem).
        timeout (float): How long to wait for shard to download before raising an exception.
    """

    def _download_file(unsigned: bool = False,
                       extra_args: Optional[dict[str, Any]] = None) -> None:
        """Download the file from AWS S3 bucket. The bucket can be either public or private.

        Args:
            unsigned (bool, optional):  Set to True if it is a public bucket.
                Defaults to ``False``.
            extra_args (Dict[str, Any], optional): Extra arguments supported by boto3.
                Defaults to ``None``.
        """
        retries = {
            'mode': 'adaptive',
        }
        if unsigned:
            # Client will be using unsigned mode in which public
            # resources can be accessed without credentials
            config = Config(read_timeout=timeout, signature_version=UNSIGNED, retries=retries)
        else:
            config = Config(read_timeout=timeout, retries=retries)

        if extra_args is None:
            extra_args = {}

        # Create a new session per thread
        session = boto3.session.Session()
        # Create a resource client using a thread's session object
        s3 = session.client('s3', config=config, endpoint_url=os.environ.get('S3_ENDPOINT_URL'))
        # Threads calling S3 operations return RuntimeError (cannot schedule new futures after
        # interpreter shutdown). Temporary solution is to have `use_threads` as `False`.
        # Issue: https://github.com/boto/boto3/issues/3113
        s3.download_file(obj.netloc,
                         obj.path.lstrip('/'),
                         local,
                         ExtraArgs=extra_args,
                         Config=TransferConfig(use_threads=False))

    import boto3
    from boto3.s3.transfer import TransferConfig
    from botocore import UNSIGNED
    from botocore.config import Config
    from botocore.exceptions import ClientError, NoCredentialsError

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 's3':
        raise ValueError(
            f'Expected obj.scheme to be `s3`, instead, got {obj.scheme} for remote={remote}')

    extra_args = {}
    # When enabled, the requester instead of the bucket owner pays the cost of the request
    # and the data download from the bucket.
    if os.environ.get('MOSAICML_STREAMING_AWS_REQUESTER_PAYS') is not None:
        requester_pays_buckets = os.environ.get(  # yapf: ignore
            'MOSAICML_STREAMING_AWS_REQUESTER_PAYS').split(',')  # pyright: ignore
        requester_pays_buckets = [name.strip() for name in requester_pays_buckets]
        if obj.netloc in requester_pays_buckets:
            extra_args['RequestPayer'] = 'requester'

    try:
        _download_file(extra_args=extra_args)
    except NoCredentialsError:
        # Public S3 buckets without credentials
        _download_file(unsigned=True, extra_args=extra_args)
    except ClientError as e:
        if e.response['Error']['Code'] in BOTOCORE_CLIENT_ERROR_CODES:
            e.args = (f'Object {remote} not found! Either check the bucket path or the bucket ' +
                      f'permission. If the bucket is a requester pays bucket, then provide the ' +
                      f'bucket name to the environment variable ' +
                      f'`MOSAICML_STREAMING_AWS_REQUESTER_PAYS`.',)
            raise e
        elif e.response['Error']['Code'] == '400':
            # Public S3 buckets without credentials
            _download_file(unsigned=True, extra_args=extra_args)
    except Exception:
        raise


def download_from_sftp(remote: str, local: str) -> None:
    """Download a file from remote SFTP server to local filepath.

    Authentication must be provided via username/password in the ``remote`` URI, or a valid SSH
    config, or a default key discoverable in ``~/.ssh/``.

    Args:
        remote (str): Remote path (SFTP).
        local (str): Local path (local filesystem).
    """
    from paramiko import SSHClient

    # Parse URL
    url = urllib.parse.urlsplit(remote)
    if url.scheme.lower() != 'sftp':
        raise ValueError('If specifying a URI, only the sftp scheme is supported.')
    if not url.hostname:
        raise ValueError('If specifying a URI, the URI must include the hostname.')
    if url.query or url.fragment:
        raise ValueError('Query and fragment parameters are not supported as part of a URI.')
    hostname = url.hostname
    port = url.port
    username = url.username
    password = url.password
    remote_path = url.path

    # Get SSH key file if specified
    key_filename = os.environ.get('COMPOSER_SFTP_KEY_FILE', None)
    known_hosts_filename = os.environ.get('COMPOSER_SFTP_KNOWN_HOSTS_FILE', None)

    # Default port
    port = port if port else 22

    # Local tmp
    local_tmp = local + '.tmp'
    if os.path.exists(local_tmp):
        os.remove(local_tmp)

    with SSHClient() as ssh_client:
        # Connect SSH Client
        ssh_client.load_system_host_keys(known_hosts_filename)
        ssh_client.connect(
            hostname=hostname,
            port=port,
            username=username,
            password=password,
            key_filename=key_filename,
        )

        # SFTP Client
        sftp_client = ssh_client.open_sftp()
        sftp_client.get(remotepath=remote_path, localpath=local_tmp)
    os.rename(local_tmp, local)


def download_from_gcs(remote: str, local: str) -> None:
    """Download a file from remote GCS to local.

    Args:
        remote (str): Remote path (GCS).
        local (str): Local path (local filesystem).
    """
    from google.auth.exceptions import DefaultCredentialsError
    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'gs':
        raise ValueError(
            f'Expected obj.scheme to be `gs`, instead, got {obj.scheme} for remote={remote}')

    if 'GCS_KEY' in os.environ and 'GCS_SECRET' in os.environ:
        _gcs_with_hmac(remote, local, obj)
    else:
        try:
            _gcs_with_service_account(local, obj)
        except (DefaultCredentialsError, EnvironmentError):
            raise ValueError(GCS_ERROR_NO_AUTHENTICATION)


def _gcs_with_hmac(remote: str, local: str, obj: urllib.parse.ParseResult) -> None:
    """Download a file from remote GCS to local using user level credentials.

    Args:
        remote (str): Remote path (GCS).
        local (str): Local path (local filesystem).
        obj (ParseResult): ParseResult object of remote.
    """
    import boto3
    from boto3.s3.transfer import TransferConfig
    from botocore.exceptions import ClientError

    # Create a new session per thread
    session = boto3.session.Session()
    # Create a resource client using a thread's session object
    gcs_client = session.client('s3',
                                region_name='auto',
                                endpoint_url='https://storage.googleapis.com',
                                aws_access_key_id=os.environ['GCS_KEY'],
                                aws_secret_access_key=os.environ['GCS_SECRET'])
    try:
        # Threads calling S3 operations return RuntimeError (cannot schedule new futures after
        # interpreter shutdown). Temporary solution is to have `use_threads` as `False`.
        # Issue: https://github.com/boto/boto3/issues/3113
        gcs_client.download_file(obj.netloc,
                                 obj.path.lstrip('/'),
                                 local,
                                 Config=TransferConfig(use_threads=False))
    except ClientError as e:
        if e.response['Error']['Code'] in BOTOCORE_CLIENT_ERROR_CODES:
            raise FileNotFoundError(f'Object {remote} not found.') from e
    except Exception:
        raise


def _gcs_with_service_account(local: str, obj: urllib.parse.ParseResult) -> None:
    """Download a file from remote GCS to local using service account credentials.

    Args:
        local (str): Local path (local filesystem).
        obj (ParseResult): ParseResult object of remote path (GCS).
    """
    from google.auth import default as default_auth
    from google.cloud.storage import Blob, Bucket, Client

    credentials, _ = default_auth()
    gcs_client = Client(credentials=credentials)
    blob = Blob(obj.path.lstrip('/'), Bucket(gcs_client, obj.netloc))
    blob.download_to_filename(local)


def download_from_oci(remote: str, local: str) -> None:
    """Download a file from remote OCI to local.

    Args:
        remote (str): Remote path (OCI).
        local (str): Local path (local filesystem).
    """
    import oci
    config = oci.config.from_file()
    client = oci.object_storage.ObjectStorageClient(
        config=config, retry_strategy=oci.retry.DEFAULT_RETRY_STRATEGY)
    namespace = client.get_namespace().data
    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'oci':
        raise ValueError(
            f'Expected obj.scheme to be `oci`, instead, got {obj.scheme} for remote={remote}')

    bucket_name = obj.netloc.split('@' + namespace)[0]
    # Remove leading and trailing forward slash from string
    object_path = obj.path.strip('/')
    object_details = client.get_object(namespace, bucket_name, object_path)
    local_tmp = local + '.tmp'
    with open(local_tmp, 'wb') as f:
        for chunk in object_details.data.raw.stream(2048**2, decode_content=False):
            f.write(chunk)
    os.rename(local_tmp, local)


def download_from_hf(remote: str, local: str) -> None:
    """Download a file from remote Hugging Face to local.

    Args:
        remote (str): Remote path (Hugging Face).
        local (str): Local path (local filesystem).
    """
    from huggingface_hub import hf_hub_download

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'hf':
        raise ValueError(f'Expected remote path to start with `hf://`, got {remote}.')

    _, _, _, repo_org, repo_name, path = remote.split('/', 5)
    local_dirname = os.path.dirname(local)
    hf_hub_download(repo_id=f'{repo_org}/{repo_name}',
                    filename=path,
                    repo_type='dataset',
                    local_dir=local_dirname)

    downloaded_name = os.path.join(local_dirname, path)
    os.rename(downloaded_name, local)


def download_from_azure(remote: str, local: str) -> None:
    """Download a file from remote Microsoft Azure to local.

    Args:
        remote (str): Remote path (azure).
        local (str): Local path (local filesystem).
    """
    from azure.core.exceptions import ResourceNotFoundError
    from azure.storage.blob import BlobServiceClient

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'azure':
        raise ValueError(
            f'Expected obj.scheme to be `azure`, instead, got {obj.scheme} for remote={remote}')

    # Create a new session per thread
    service = BlobServiceClient(
        account_url=f"https://{os.environ['AZURE_ACCOUNT_NAME']}.blob.core.windows.net",
        credential=os.environ['AZURE_ACCOUNT_ACCESS_KEY'])
    try:
        file_path = obj.path.lstrip('/').split('/')
        container_name = file_path[0]
        blob_name = os.path.join(*file_path[1:])
        blob_client = service.get_blob_client(container=container_name, blob=blob_name)
        local_tmp = local + '.tmp'
        with open(local_tmp, 'wb') as my_blob:
            blob_data = blob_client.download_blob()
            blob_data.readinto(my_blob)
        os.rename(local_tmp, local)
    except ResourceNotFoundError as e:
        raise FileNotFoundError(f'Object {remote} not found.') from e
    except Exception:
        raise


def download_from_azure_datalake(remote: str, local: str) -> None:
    """Download a file from remote Microsoft Azure Data Lake to local.

    Args:
        remote (str): Remote path (azure).
        local (str): Local path (local filesystem).
    """
    from azure.core.exceptions import ResourceNotFoundError
    from azure.storage.filedatalake import DataLakeServiceClient

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'azure-dl':
        raise ValueError(
            f'Expected obj.scheme to be `azure-dl`, got {obj.scheme} for remote={remote}')

    # Create a new session per thread
    service = DataLakeServiceClient(
        account_url=f"https://{os.environ['AZURE_ACCOUNT_NAME']}.dfs.core.windows.net",
        credential=os.environ['AZURE_ACCOUNT_ACCESS_KEY'],
    )
    try:
        file_client = service.get_file_client(file_system=obj.netloc,
                                              file_path=obj.path.lstrip('/'))
        local_tmp = local + '.tmp'
        with open(local_tmp, 'wb') as my_file:
            file_data = file_client.download_file()
            file_data.readinto(my_file)
        os.rename(local_tmp, local)
    except ResourceNotFoundError as e:
        raise FileNotFoundError(f'Object {remote} not found.') from e
    except Exception:
        raise


def download_from_databricks_unity_catalog(remote: str, local: str) -> None:
    """Download a file from remote Databricks Unity Catalog to local.

    .. note::
        The Databricks UC Volume path must be of the form
        `dbfs:/Volumes/<catalog-name>/<schema-name>/<volume-name>/path`.

    Args:
        remote (str): Remote path (Databricks Unity Catalog).
        local (str): Local path (local filesystem).
    """
    try:
        from databricks.sdk import WorkspaceClient
        from databricks.sdk.core import DatabricksError
    except ImportError as e:
        e.msg = get_import_exception_message(e.name, 'databricks')  # pyright: ignore
        raise e

    path = pathlib.Path(remote)
    provider_prefix = os.path.join(path.parts[0], path.parts[1])
    if provider_prefix != 'dbfs:/Volumes':
        raise ValueError(
            f'Expected path prefix to be `dbfs:/Volumes` if it is a Databricks Unity Catalog, ' +
            f'instead, got {provider_prefix} for remote={remote}.')

    client = WorkspaceClient()
    file_path = urllib.parse.urlparse(remote)
    local_tmp = local + '.tmp'
    try:
        with client.files.download(file_path.path).contents as response:
            with open(local_tmp, 'wb') as f:
                # Download data in chunks to avoid memory issues.
                for chunk in iter(lambda: response.read(64 * 1024 * 1024), b''):
                    f.write(chunk)
    except DatabricksError as e:
        if e.error_code == 'REQUEST_LIMIT_EXCEEDED':
            e.args = (f'Dataset download request has been rejected due to too many concurrent ' +
                      f'operations. Increase the `download_retry` value to retry downloading ' +
                      f'a file.',)
        if e.error_code == 'NOT_FOUND':
            raise FileNotFoundError(f'Object {remote} not found.') from e
        raise e
    os.rename(local_tmp, local)


def download_from_dbfs(remote: str, local: str) -> None:
    """Download a file from remote Databricks File System to local.

    Args:
        remote (str): Remote path (dbfs).
        local (str): Local path (local filesystem).
    """
    try:
        from databricks.sdk import WorkspaceClient
        from databricks.sdk.core import DatabricksError
    except ImportError as e:
        e.msg = get_import_exception_message(e.name, 'databricks')  # pyright: ignore
        raise e

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'dbfs':
        raise ValueError(f'Expected path prefix to be `dbfs` if it is a Databricks File System, ' +
                         f'instead, got {obj.scheme} for remote={remote}.')

    client = WorkspaceClient()
    file_path = urllib.parse.urlparse(remote)
    local_tmp = local + '.tmp'
    try:
        with client.dbfs.download(file_path.path) as response:
            with open(local_tmp, 'wb') as f:
                # Multiple shard files are getting downloaded in parallel, so we need to
                # read the data in chunks to avoid memory issues.
                # Read 1MB of data at a time since that's the max limit it can read at a time.
                for chunk in iter(lambda: response.read(1024 * 1024), b''):
                    f.write(chunk)
    except DatabricksError as e:
        if e.error_code == 'PERMISSION_DENIED':
            e.args = (f'Ensure the file path or credentials are set correctly. For ' +
                      f'Databricks Unity Catalog, file path must starts with `dbfs:/Volumes` ' +
                      f'and for Databricks File System, file path must starts with `dbfs`. ' +
                      e.args[0],)
        raise e
    os.rename(local_tmp, local)


def download_from_alipan(remote: str, local: str) -> None:
    """Download a file from remote Alipan to local.

    Args:
        remote (str): Remote path (Alipan).
        local (str): Local path (local filesystem).
    """
    from alipcs_py.alipcs import AliPCSApiMix
    from alipcs_py.commands.download import download_file

    web_refresh_token = os.environ['ALIPAN_WEB_REFRESH_TOKEN']
    web_token_type = 'Bearer'
    alipan_encrypt_password = os.environ.get('ALIPAN_ENCRYPT_PASSWORD', '').encode()

    api = AliPCSApiMix(web_refresh_token, web_token_type=web_token_type)

    obj = urllib.parse.urlparse(remote)
    if obj.scheme != 'alipan':
        raise ValueError(
            f'Expected obj.scheme to be `alipan`, instead, got {obj.scheme} for remote={remote}')
    if obj.netloc != '':
        raise ValueError(
            f'Expected remote to be alipan:///path/to/some, instead, got remote={remote}')

    remote_path = obj.path
    filename = pathlib.PosixPath(remote_path).name
    localdir = pathlib.Path(local).parent

    remote_pcs_file = api.get_file(remotepath=remote_path)
    if remote_pcs_file is None:
        raise FileNotFoundError(f'Object {remote} not found.')

    download_file(
        api,
        remote_pcs_file,
        localdir=localdir,
        downloader='me',
        concurrency=1,
        show_progress=False,
        encrypt_password=alipan_encrypt_password,
    )
    os.rename(localdir / filename, local)


def download_from_local(remote: str, local: str) -> None:
    """Download a file from remote to local.

    Args:
        remote (str): Remote path (local filesystem).
        local (str): Local path (local filesystem).
    """
    local_tmp = local + '.tmp'
    if os.path.exists(local_tmp):
        os.remove(local_tmp)
    shutil.copy(remote, local_tmp)
    os.rename(local_tmp, local)


def download_file(remote: Optional[str], local: str, timeout: float):
    """Use the correct download handler to download the file.

    Args:
        remote (str, optional): Remote path (local filesystem).
        local (str): Local path (local filesystem).
        timeout (float): How long to wait for file to download before raising an exception.
    """
    if os.path.exists(local):
        return

    # fix paths for windows
    local = local.replace('\\', '/')
    if remote:
        remote = remote.replace('\\', '/')

    local_dir = os.path.dirname(local)
    os.makedirs(local_dir, exist_ok=True)

    if not remote:
        raise ValueError(
            'In the absence of local dataset, path to remote dataset must be provided')
    elif remote.startswith('s3://'):
        download_from_s3(remote, local, timeout)
    elif remote.startswith('sftp://'):
        download_from_sftp(remote, local)
    elif remote.startswith('gs://'):
        download_from_gcs(remote, local)
    elif remote.startswith('oci://'):
        download_from_oci(remote, local)
    elif remote.startswith('hf://'):
        download_from_hf(remote, local)
    elif remote.startswith('azure://'):
        download_from_azure(remote, local)
    elif remote.startswith('azure-dl://'):
        download_from_azure_datalake(remote, local)
    elif remote.startswith('dbfs:/Volumes'):
        download_from_databricks_unity_catalog(remote, local)
    elif remote.startswith('dbfs:/'):
        download_from_dbfs(remote, local)
    elif remote.startswith('alipan://'):
        download_from_alipan(remote, local)
    else:
        download_from_local(remote, local)


def wait_for_download(local: str, timeout: float = 60) -> None:
    """Wait for a download by another thread/process to complete.

    Args:
        local (str): Local path (local filesystem).
        timeout (float): How long to wait for file to download before raising an exception.
            Defaults to ``60``.
    """
    t0 = time()
    while not os.path.exists(local):
        elapsed = time() - t0
        if timeout < elapsed:
            raise TimeoutError(
                f'Waited longer than {timeout}s for other worker to download {local}.')
        sleep(0.25)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Simulator web UI using streamlit."""

import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import math
from concurrent.futures import ProcessPoolExecutor
from io import StringIO
from typing import Union

import numpy as np
import pandas as pd
import streamlit as st
import yaml
from core.create_index import create_stream_index
from core.main import simulate
from core.shuffle_quality import analyze_shuffle_quality_entropy
from core.sim_dataset import SimulationDataset
from core.sim_time import Time
from core.utils import get_total_batches
from core.yaml_processing import create_simulation_dataset, ingest_yaml
from interfaces.interface_utils import get_train_dataset_params
from interfaces.widgets import (display_shuffle_quality_graph, display_simulation_stats,
                                get_line_chart, param_inputs)

from streaming.base.util import bytes_to_int, number_abbrev_to_int

# set up page
st.set_page_config(layout='wide')
col1, space, col2 = st.columns((10, 1, 6))
col2.title('Streaming Simulator')
col2.write('Enter run parameters in the left panel.')
col2.text('')
progress_bar = col1.progress(0)
status_text = col1.empty()
col1.text('')
throughput_plot = col2.empty()
network_plot = col2.empty()
sim_stats = col2.empty()
col2.text('')
shuffle_quality_plot = col2.empty()
throughput_window = 10
shuffle_quality_algos = ['naive', 'py1b', 'py1br', 'py1e', 'py1s', 'py2s', 'none']


def submit_jobs(shuffle_quality: bool, dataset: SimulationDataset, time_per_sample: float,
                node_internet_bandwidth: Union[int, str], max_duration: Time):
    """Submit jobs to the executor for simulation.

    Args:
        shuffle_quality (bool): Whether to run shuffle quality analysis.
        dataset (SimulationDataset): Dataset to use for simulation.
        time_per_sample (float): Time for one device to process one sample.
        node_internet_bandwidth (Union[int,str]): Internet bandwidth per node.
        max_duration (Time): Maximum duration of simulation.
    """
    total_batches = get_total_batches(dataset=dataset, max_duration=max_duration)
    node_internet_bandwidth = bytes_to_int(node_internet_bandwidth)
    cache_limit = dataset.get_cache_limit()
    gen_sim = simulate(dataset,
                       time_per_sample,
                       node_internet_bandwidth,
                       max_duration,
                       generator=True)
    gen_step_times = []
    gen_step_downloads = []
    rolling_throughput_data = []
    immediate_throughput_data = []
    network_data = []
    steps = []
    time_to_first_batch = 0
    futures = []
    shuffle_quality_graphed = False
    # Initialize min_cache_limit to be 0. Will be replaced by the simulated value.
    min_cache_limit = 0
    # Define partial function to pass to executor map for simulation.
    with ProcessPoolExecutor() as executor:
        # Submit shuffle quality job to executor.
        if shuffle_quality:
            col1.write('Starting shuffle quality analysis...')
            input_params = st.session_state['input_params']
            # Use multiprocessing to get the shuffle quality results.
            canonical_nodes = input_params['canonical_nodes']
            if canonical_nodes is None:
                canonical_nodes = dataset.get_num_canonical_nodes()
            physical_nodes = input_params['physical_nodes']
            devices = input_params['devices']
            workers = input_params['workers']
            device_batch_size = input_params['device_batch_size']
            shuffle_block_size = number_abbrev_to_int(input_params['shuffle_block_size']) \
                if input_params['shuffle_block_size'] is not None \
                    else dataset.get_shuffle_block_size()
            samples_per_shard = dataset.get_avg_samples_per_shard()
            epoch_size = dataset.get_epoch_size()
            if epoch_size > 100_000_000:
                st.warning('Epoch size is over 100 million samples. Shuffle quality analysis \
                           will be conducted only on the first 100 million samples.',
                           icon='⚠️')
            seed = input_params['seed']
            # Submit all shuffle quality analysis jobs to executor.
            futures = [
                executor.submit(analyze_shuffle_quality_entropy, algo, canonical_nodes,
                                physical_nodes, devices, workers, device_batch_size,
                                shuffle_block_size, samples_per_shard, epoch_size, seed)
                for algo in shuffle_quality_algos
            ]

        # Simulate only on the main worker, otherwise it's super slow.
        for output in gen_sim:
            # If output is a length 2, it is the time to first batch and min cache limit.
            # Otherwise it is the step, step time, and shard download from the simulation.
            if len(output) == 2:
                step = total_batches - 1
                time_to_first_batch, min_cache_limit = output
            else:
                if len(output) != 3:
                    raise ValueError(f'Simulation with generate=True should return 3 results per' +
                                     f' step. Instead, output was length {len(output)}.')

                step, step_time, shard_download = output
                gen_step_times.append(step_time)
                gen_step_downloads.append(shard_download)
                # plot throughput once we have enough samples for the window
                rolling_throughput = 0
                if step >= throughput_window - 1:
                    step_time_window = np.array(gen_step_times[-throughput_window:])
                    rolling_throughput = 1 / np.mean((step_time_window))
                rolling_throughput_data.append(rolling_throughput)
                immediate_throughput_data.append(1 / step_time)
                # plot network usage
                cumulative_shard_download = np.sum(np.array(gen_step_downloads))
                network_data.append(cumulative_shard_download)
                steps.append(step + 1)

            # update plots and percentages at regular intervals
            plot_interval = math.ceil(total_batches / 15)
            if step == 1 or step % plot_interval == 0 or step == total_batches - 1:
                rolling_throughput_df = pd.DataFrame({
                    'step': steps,
                    'measurement': [' rolling avg'] * len(rolling_throughput_data),
                    'throughput (batches/s)': rolling_throughput_data
                })
                throughput_df = rolling_throughput_df
                network_df = pd.DataFrame({
                    'step': steps,
                    'cumulative network usage (bytes)': network_data
                })
                throughput_plot.altair_chart(get_line_chart(throughput_df, throughput_window,
                                                            True),
                                             use_container_width=True)
                network_plot.altair_chart(get_line_chart(network_df, throughput_window, False),
                                          use_container_width=True)
                # update progress bar and text
                percentage = int(100 * (step + 1) / (total_batches))
                status_text.text('%i%% Complete' % percentage)
                progress_bar.progress(percentage)

                # If applicable, check if the shuffle quality tasks are finished, and graph.
                if shuffle_quality and all(f.done() for f in futures) \
                    and not shuffle_quality_graphed:
                    display_shuffle_quality_graph(futures, shuffle_quality_plot)
                    shuffle_quality_graphed = True

        gen_step_times = np.array(gen_step_times)
        gen_step_downloads = np.array(gen_step_downloads)
        device_batch_size = dataset.get_batch_size()
        display_simulation_stats(sim_stats, total_batches, gen_step_times, time_per_sample,
                                 device_batch_size, time_to_first_batch, min_cache_limit,
                                 cache_limit)

        # If shuffle quality still hasn't been graphed yet, we get the result and graph it.
        if shuffle_quality and not shuffle_quality_graphed:
            display_shuffle_quality_graph(futures, shuffle_quality_plot)
            shuffle_quality_graphed = True


def get_input_params_initial(physical_nodes: int, devices: int, workers: int,
                             global_batch_size: int, train_dataset: dict, max_duration: Time,
                             time_per_sample: float, node_internet_bandwidth: Union[int, str]):
    """Create input_params and dataset for the first time.

    This function is called when modify_params is clicked, or when the run is submitted for
    simulation when using a yaml file.

    Args:
        physical_nodes (int): Number of physical nodes.
        devices (int): Number of devices per node.
        workers (int): Number of workers.
        global_batch_size (int): Global batch size.
        train_dataset (dict): Train dataset parameters.
        max_duration (Time): Maximum duration of simulation.
        time_per_sample (float): Time for one device to process one sample.
        node_internet_bandwidth (Union[int,str]): Internet bandwidth per node.
    """
    try:
        st.session_state['creating_dataset'] = True
        dataset = create_simulation_dataset(physical_nodes, devices, workers, global_batch_size,
                                            train_dataset)
        st.session_state['orig_dataset'] = dataset
        input_params = {}
        # dataset input_params
        input_params['streams'] = dataset.get_stream_info()
        # training input_params
        input_params['max_duration'] = max_duration
        input_params['epoch_size'] = dataset.get_epoch_size()
        input_params['device_batch_size'] = dataset.get_batch_size()
        # hardware and network input_params
        input_params['physical_nodes'] = physical_nodes
        input_params['devices'] = devices
        input_params['time_per_sample'] = time_per_sample
        input_params['node_network_bandwidth'] = node_internet_bandwidth
        # streaming input_params
        input_params['workers'] = workers
        input_params['canonical_nodes'] = dataset.get_num_canonical_nodes()
        input_params['predownload'] = dataset.get_predownload()
        input_params['shuffle'] = dataset.get_shuffle()
        input_params['shuffle_algo'] = dataset.get_shuffle_algo()
        input_params['shuffle_block_size'] = dataset.get_shuffle_block_size()
        input_params['seed'] = dataset.get_shuffle_seed()
        input_params['cache_limit'] = dataset.get_cache_limit()
        input_params['sampling_method'] = dataset.get_sampling_method()
        input_params['sampling_granularity'] = dataset.get_sampling_granularity()
        input_params['batching_method'] = dataset.get_batching_method()
        # Save input_params and originally created dataset to session state.
        st.session_state['input_params'] = input_params
    except FileNotFoundError:
        st.error('Please wait until the dataset is loaded before changing toggle values too \
                 quickly. Doing so can cause issues with creating multiple datasets, since \
                 Streamlit reloads widgets every single time a toggle value changes.',
                 icon='🚨')


# Define parameter input area.

# Check if the user wants to submit a yaml file.
use_yaml = col1.toggle(':sparkles: **Use `yaml`** :sparkles:', value=True)

if use_yaml:
    uploaded_yaml = col1.file_uploader('Upload a yaml file', type=['yaml'])
    if uploaded_yaml is not None:
        string_yaml = StringIO(uploaded_yaml.getvalue().decode('utf-8')).read()
        dict_yaml = yaml.safe_load(string_yaml)
        total_devices, workers, max_duration, global_batch_size, train_dataset = \
        ingest_yaml(yaml_dict=dict_yaml)
        # Check which parameters we still need to ask for.
        col1.write('The parameters below were not found in your yaml file. Enter them here:')
        physical_nodes = col1.number_input(
            'number of physical nodes',
            step=1,
            value=total_devices // 8 if total_devices is not None else 1,
            help=
            'number of physical nodes for this run. a node typically consists of 8 devices (GPUs).'
        )
        physical_nodes = int(physical_nodes)
        # Using physical_nodes, calculate number of devices per node.
        if total_devices is None:
            devices = col1.number_input(
                'devices per node',
                step=1,
                value=8,
                help=
                'number of devices (GPUs) per node for this run. there are typically 8 devices per node.'
            )
        else:
            if total_devices % physical_nodes != 0:
                raise ValueError('The number of devices must be divisible by the number of nodes.')
            devices = total_devices // physical_nodes
        devices = int(devices)
        time_per_sample = col1.number_input(
            'process time per sample (s)',
            step=0.0005,
            value=0.0175,
            format='%.4f',
            help='time for one device to process one sample from your dataset.')
        time_per_sample = float(time_per_sample)
        node_internet_bandwidth = col1.text_input('network bandwidth per node (bytes/s)',
                                                  value='500MB',
                                                  help='network bandwidth available to each \
                                                node. in practice, network bandwidth is \
                                                variable and is affected by many factors, \
                                                including cluster demand.')
        col1.warning(
            'If you are using a subclass of StreamingDataset with custom defaults, please \
                   select "Modify Parameters" and verify that the input parameters are correct.',
            icon='⚠️')
        submitted = col1.button('Simulate Run', use_container_width=True)
        shuffle_quality = col1.toggle('Analyze Shuffle Quality',
                                      value=False,
                                      help='Analyze shuffle qualities for this run for different \
                                        shuffle algos using an entropy-based metric. ⚠️ **Results \
                                        are *noisy estimates* and may not reflect the true \
                                        shuffle quality.**')
        modify_params = col1.toggle('Modify Parameters', value=False)

        # Display components and take actions based on the values of the above three buttons.
        if modify_params:
            # Create dataset and input_params if it doesn't already exist.
            if 'input_params' not in st.session_state:
                col1.write('Preparing dataset for modification...')
                get_input_params_initial(physical_nodes, devices, workers, global_batch_size,
                                         train_dataset, max_duration, time_per_sample,
                                         node_internet_bandwidth)
            # We have input_params in the session state. Use it to populate the form.
            defaults = st.session_state['input_params']
            # Define parameter input area with default values.
            input_params = {}
            param_inputs(col1, input_params, defaults=defaults)
            # input_params has been repopulated with new values. Save to session state.
            st.session_state['input_params'] = input_params

        if submitted:
            # Create dataset if it is not yet present.
            if 'input_params' not in st.session_state:
                col1.write('Preparing dataset for this run...')
                get_input_params_initial(physical_nodes, devices, workers, global_batch_size,
                                         train_dataset, max_duration, time_per_sample,
                                         node_internet_bandwidth)
            # If modify_params is false, we submit the jobs using the original dataset from yaml.
            if not modify_params:
                col1.write('Starting Simulation...')
                dataset = st.session_state['orig_dataset']
                # shuffle_quality is passed through to the job submission function.
                submit_jobs(shuffle_quality, dataset, time_per_sample, node_internet_bandwidth,
                            max_duration)
            else:
                # If modify_params is true, we retrieve the most recent input params from session
                # state, create a new dataset, and submit the jobs.
                col1.write('Preparing dataset with modifications...')
                # Get parameters for new SimulationDataset from input_params and train_dataset.
                input_params = st.session_state['input_params']
                train_dataset = get_train_dataset_params(input_params, old_params=train_dataset)
                # Get the rest of the needed params from the new inputs
                physical_nodes = input_params['physical_nodes']
                devices = input_params['devices']
                global_batch_size = input_params['device_batch_size'] * devices * physical_nodes
                workers = input_params['workers']
                max_duration = input_params['max_duration']
                time_per_sample = input_params['time_per_sample']
                node_internet_bandwidth = input_params['node_network_bandwidth']
                # Make sure node_internet_bandwidth is an int.
                dataset = create_simulation_dataset(physical_nodes, devices, workers,
                                                    global_batch_size, train_dataset)
                col1.write('Starting Simulation...')
                submit_jobs(shuffle_quality, dataset, time_per_sample, node_internet_bandwidth,
                            max_duration)
    else:
        # In this case, no file is uploaded, and we should clear dataset and input params if needed
        if 'input_params' in st.session_state:
            del st.session_state['input_params']
        if 'orig_dataset' in st.session_state:
            del st.session_state['orig_dataset']
else:
    submitted = col1.button('Simulate Run', use_container_width=True)
    col1.text('')
    shuffle_quality = col1.toggle('Analyze Shuffle Quality',
                                  value=False,
                                  help='Analyze shuffle qualities for this run for different \
                                        shuffle algos using an entropy-based metric. ⚠️ **Results \
                                        are *noisy estimates* and may not reflect the true \
                                        shuffle quality.**')
    if 'input_params' in st.session_state:
        st.session_state['input_params'] = {}
    input_params = {}
    param_inputs(col1, input_params, defaults=input_params)
    if submitted:
        # Params have been submitted. Create new dataset and proceed with simulation.
        col1.write('Preparing dataset for this run...')
        # Create index files and Stream object for each stream.
        streams = {}
        for stream_idx, stream in input_params['streams'].items():
            stream_dict = {}
            if 'path' in stream:
                # Case when user has provided a path to an index.json file.
                stream_folder = os.path.dirname(stream['path'])
                if stream['path_type'] == 'local':
                    stream_dict['local'] = stream_folder
                else:
                    stream_dict['remote'] = stream_folder
            else:
                # Case when user provides estimates for stream characteristics.
                index_path = create_stream_index(stream['shards'], stream['samples_per_shard'],
                                                 stream['avg_raw_shard_size'],
                                                 stream['avg_zip_shard_size'])
                stream_folder = os.path.dirname(index_path)
                stream_dict['local'] = stream_folder
            stream_dict['proportion'] = stream['proportion']
            stream_dict['repeat'] = stream['repeat']
            stream_dict['choose'] = stream['choose']
            streams[stream_idx] = stream_dict
        input_params['streams'] = streams
        # Get parameters for new SimulationDataset from input_params and train_dataset.
        train_dataset = get_train_dataset_params(input_params)
        # Get the rest of the needed params from the new inputs
        physical_nodes = input_params['physical_nodes']
        devices = input_params['devices']
        global_batch_size = input_params['device_batch_size'] * devices * physical_nodes
        workers = input_params['workers']
        max_duration = input_params['max_duration']
        time_per_sample = input_params['time_per_sample']
        node_internet_bandwidth = input_params['node_network_bandwidth']
        dataset = create_simulation_dataset(physical_nodes, devices, workers, global_batch_size,
                                            train_dataset)
        # Make sure input_params is in session state.
        st.session_state['input_params'] = input_params
        col1.write('Starting Simulation...')
        submit_jobs(shuffle_quality, dataset, time_per_sample, node_internet_bandwidth,
                    max_duration)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Peripheral functions for interface functionality."""

import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from typing import Optional

import numpy as np
from core.utils import get_rolling_avg_throughput
from numpy.typing import NDArray

from streaming.base.util import number_abbrev_to_int


def plot_simulation(step_times: NDArray, step_downloads: NDArray, window: int = 10):
    """Plots simulation results for web UI or local script.

    Args:
        step_times (NDArray): time per step, as calculated by simulation
        step_downloads (NDArray): download size (bytes) per step, as calculated by simulation
        window (int, optional): window size to calculate batch throughput over. Defaults to ``10``.
    """
    import matplotlib.pyplot as plt

    immediate_batch_throughput = 1 / step_times

    step_downloads_cumulative = np.cumsum(step_downloads)

    batch_throughput_rolling_avg = get_rolling_avg_throughput(step_times, window)

    # matplotlib plot with 2 vertically stacked subplots
    fig, (ax1, ax2) = plt.subplots(2, 1)

    plt.suptitle('Simulation Results', fontsize=16)

    ax1.plot(np.arange(immediate_batch_throughput.shape[0]),
             immediate_batch_throughput,
             color='lightblue',
             label='per step throughput')
    ax1.plot(np.arange(batch_throughput_rolling_avg.shape[0]),
             batch_throughput_rolling_avg,
             color='darkblue',
             label='rolling throughput (10 step avg)')
    ax1.legend()
    ax1.set_ylim([0, max(immediate_batch_throughput) * 1.1])
    ax1.set_ylabel('batches/s')
    ax1.set_title('batch throughput (batches/s)')

    ax2.plot(np.arange(step_downloads_cumulative.shape[0]),
             step_downloads_cumulative,
             color='blue',
             label='total')
    ax2.set_ylim([0, max(step_downloads_cumulative) * 1.1])
    ax2.set_xlabel('step')
    ax2.set_ylabel('cumulative download (bytes)')
    ax2.set_title('network traffic (bytes)')

    fig.set_figheight(8)
    fig.set_figwidth(6)

    plt.show()


def get_train_dataset_params(input_params: dict, old_params: Optional[dict] = None) -> dict:
    """Get train dataset params from input params.

    Args:
        input_params (dict): The input parameter dictionary set by the user.
        old_params (Optional[dict], optional): Old parameters that may have been read in.
            Defaults to ``None``.

    Returns:
        dict: The full train dataset parameters.
    """
    train_dataset_params = {}
    train_dataset_params['epoch_size'] = input_params['epoch_size']
    train_dataset_params['batch_size'] = input_params['device_batch_size']
    train_dataset_params['nodes'] = input_params['physical_nodes']
    train_dataset_params['devices'] = input_params['devices']
    train_dataset_params['workers'] = input_params['workers']
    train_dataset_params['num_canonical_nodes'] = input_params['canonical_nodes']
    train_dataset_params['predownload'] = input_params['predownload']
    train_dataset_params['cache_limit'] = input_params['cache_limit']
    train_dataset_params['shuffle'] = input_params['shuffle']
    train_dataset_params['shuffle_algo'] = input_params['shuffle_algo']
    train_dataset_params['shuffle_block_size'] = number_abbrev_to_int(
        input_params['shuffle_block_size']) if input_params['shuffle_block_size'] is not None \
        else None
    train_dataset_params['shuffle_seed'] = input_params['seed']
    train_dataset_params['sampling_method'] = input_params['sampling_method']
    train_dataset_params['sampling_granularity'] = input_params['sampling_granularity']
    train_dataset_params['batching_method'] = input_params['batching_method']

    # If there were old params, fill them in.
    if old_params is not None:
        existing_params_set = set(train_dataset_params.keys())
        old_params_set = set(old_params.keys())
        # Keep params that were set in yaml but not accessible by the user in the UI.
        # This includes the old_params "local"/"remote" or "streams".
        for param in old_params_set - existing_params_set:
            train_dataset_params[param] = old_params[param]
    else:
        # If there are no old params, we need to set streams to what the user provided.
        train_dataset_params['streams'] = input_params['streams']

    return train_dataset_params



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Download videos, saving subsets as MDS datasets, then iterate over them together.

Instructions:

1. Navigate to the download section of https://m-bain.github.io/webvid-dataset/, where you will
   find 2.5M and 10M dataset splits:

   2.5M:
   - train: http://www.robots.ox.ac.uk/~maxbain/webvid/results_2M_train.csv (640MB)
   - val: http://www.robots.ox.ac.uk/~maxbain/webvid/results_2M_val.csv (1.3MB)

   10M:
   - train: http://www.robots.ox.ac.uk/~maxbain/webvid/results_10M_train.csv (2.7GB)
   - val: http://www.robots.ox.ac.uk/~maxbain/webvid/results_10M_val.csv (1.3MB)

2. Download each CSV you want to process.

3. Run this script with flags --csv (CSV) --mds_root (MDS root)
"""

import csv
import os
import re
from argparse import ArgumentParser, Namespace
from multiprocessing import Pool
from typing import Any, Iterator, Optional

import requests

from streaming import MDSWriter, Stream, StreamingDataset

# For parsing the duration field.
duration_pattern = re.compile('^PT\\d{2}H\\d{2}M\\d{2}S$')
digits_pattern = re.compile('\\d{2}')


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Returns:
        Command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--csv',
        type=str,
        required=True,
        help='Dataset CSV file from https://m-bain.github.io/webvid-dataset/',
    )
    args.add_argument(
        '--filters',
        type=str,
        default='plane,train,automobile,cat,dog',
        help='Comma-separated list of keywords to filter into sub-datasets',
    )
    args.add_argument(
        '--mds_root',
        type=str,
        required=True,
        help='Root directory path to store the output datasets',
    )
    args.add_argument(
        '--num_procs',
        type=int,
        default=64,
        help='Number of processes to use for downloading videos',
    )
    args.add_argument(
        '--limit',
        type=int,
        default=-1,
        help='Only process the first "limit" number of samples, or all of them if set to -1',
    )
    return args.parse_args()


def get_matches(filters: list[str], text: str) -> list[str]:
    """Get whether one or more filters match the given text.

    Args:
        filters (List[str]): List of substrings to match.
        text (str): Text to search for matches.

    Returns:
        List[str]: List of filters that matched.
    """
    matches = []
    for substr in filters:
        if substr in text:
            matches.append(substr)
    return matches


def each_todo(filename: str, filters: list[str]) -> Iterator[dict[str, Any]]:
    """Get each sample to download.

    Args:
        filename (str): Path to CSV file containing samples to download.

    Returns:
        Iterator[Dict[str, Any]]: Each sample to download.
    """
    it = csv.reader(open(filename))
    keys = next(it)
    for values in it:
        obj = dict(zip(keys, values))
        matches = get_matches(filters, obj['name'])
        if matches:
            obj['matches'] = matches  # pyright: ignore
            yield obj


def head(items: Iterator, limit: int) -> Iterator:
    """Take the first "limit" number of items from an iterator.

    Args:
        items (Iterator): The iterator over the items.
        limit (int): Maximum number of items to return.

    Returns:
        Iterator: An iterator over the first "limit" items.
    """
    for i, item in enumerate(items):
        if i == limit:
            return
        yield item


def parse_duration(text: str) -> int:
    """Parse a duration string into seconds.

    Args:
        text (str): Duration string.

    Returns:
        int: Duration in seconds.
    """
    assert duration_pattern.match(text)
    hours, minutes, seconds = map(int, digits_pattern.findall(text))
    return hours * 3600 + minutes * 60 + seconds


def download(obj: dict[str, Any]) -> Optional[dict[str, Any]]:
    """Download the video for the given sample.

    Args:
        obj (Dict[str, Any]): Sample to download.

    Returns:
        Optional[Dict[str, Any]]: Downloaded sample, or None if download failed.
    """
    url = obj['contentUrl']
    try:
        ret = requests.get(url)
    except:
        return None
    if ret.status_code != 200:
        return None
    obj['duration'] = parse_duration(obj['duration'])
    obj['content'] = ret.content
    return obj


def main(args: Namespace) -> None:
    """Download videos, creating an MDS dataset.

    Args:
        args (Namespace): Command-line arguments.
    """
    # Get the list of strings for filtering into sub-datasets.
    filters = args.filters.split(',') if args.filters else []

    # Define the dataset schema.
    columns = {
        'videoid': 'str',
        'name': 'str',
        'page_dir': 'str',
        'duration': 'int',
        'contentUrl': 'str',
        'content': 'bytes',
    }

    # Get each sample to crawl.
    todos = each_todo(args.csv, filters)
    if args.limit:
        todos = head(todos, args.limit)

    # Crawl samples in parallel.
    pool = Pool(args.num_procs)
    writers = {}
    for sample in pool.imap_unordered(download, todos):
        if not sample:
            continue
        for match in sample['matches']:
            writer = writers.get(match)
            if writer is None:
                dirname = os.path.join(args.mds_root, match)
                writers[match] = writer = MDSWriter(out=dirname, columns=columns)
            writer.write(sample)
    for name in sorted(filters):
        writers[name].finish()

    # Did we crawl enough?
    subsets_present = []
    for substr in filters:
        dirname = os.path.join(args.mds_root, substr)
        subsets_present.append(dirname)

    # Create a Stream per sub-dataset, then pass them to a StreamingDataset.
    streams = []
    for name in sorted(subsets_present):
        dirname = os.path.join(args.mds_root, name)
        stream = Stream(local=dirname, proportion=1 / len(subsets_present))
        streams.append(stream)
    dataset = StreamingDataset(streams=streams, epoch_size=50, batch_size=1)

    # Print the size of each sub-dataset.
    for name, num_samples in zip(sorted(subsets_present), dataset.samples_per_stream):
        print(f'Subset "{name}": {num_samples} samples')

    # Iterate the combined dataset 3 times.
    for epoch in range(3):
        print(f'Epoch {epoch}:')
        for idx, sample in enumerate(dataset):
            text = ' '.join(sample['name'].split())
            print(f'{idx:6} | {text}')


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest

from streaming.base.spanner import Spanner


def test_spanner_success():
    shard_sizes = np.arange(5, 100, 5)
    span_size = 7
    spanner = Spanner(shard_sizes, span_size)
    index = 0
    for wanted_shard_id, shard_size in enumerate(shard_sizes):
        for wanted_offset in range(shard_size):
            got_shard_id, got_offset = spanner[index]
            assert got_shard_id == wanted_shard_id
            assert got_offset == wanted_offset
            index += 1


@pytest.mark.parametrize('index', [-10, 2000])
def test_spanner_invalid_index(index: int):
    shard_sizes = np.arange(5, 100, 5)
    span_size = 7
    with pytest.raises(IndexError, match='Invalid sample index.*'):
        spanner = Spanner(shard_sizes, span_size)
        spanner[index]



"""Create masked LM/next sentence masked_lm TF examples for BERT."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import numpy as np
import random
import tokenization
import tensorflow.compat.v1 as tf

from streaming import MDSWriter

flags = tf.flags

FLAGS = flags.FLAGS

flags.DEFINE_string("input_file", None,
                    "Input raw text file (or comma-separated list of files).")

flags.DEFINE_string("output_dir", None, "Output example directory.")

flags.DEFINE_string("compression", "zstd:7",
                    "Optional compression algorithm and compression level.")

flags.DEFINE_string("hashes", "", "Optional list of hashing algorithms with " +
                    "which to validate produced shards.")

flags.DEFINE_integer("size_limit", 1 << 26,
                     "Optional shard size limit in bytes")

flags.DEFINE_string("vocab_file", None,
                    "The vocabulary file that the BERT model was trained on.")

flags.DEFINE_bool(
    "do_lower_case", True,
    "Whether to lower case the input text. Should be True for uncased "
    "models and False for cased models.")

flags.DEFINE_integer("max_seq_length", 128, "Maximum sequence length.")

flags.DEFINE_integer("max_predictions_per_seq", 20,
                     "Maximum number of masked LM predictions per sequence.")

flags.DEFINE_integer("random_seed", 12345, "Random seed for data generation.")

flags.DEFINE_integer(
    "dupe_factor", 10,
    "Number of times to duplicate the input data (with different masks).")

flags.DEFINE_float("masked_lm_prob", 0.15, "Masked LM probability.")

flags.DEFINE_float(
    "short_seq_prob", 0.1,
    "Probability of creating sequences which are shorter than the "
    "maximum length.")


class TrainingInstance(object):
  """A single training instance (sentence pair)."""

  def __init__(self, tokens, segment_ids, masked_lm_positions, masked_lm_labels,
               is_random_next):
    self.tokens = tokens
    self.segment_ids = segment_ids
    self.is_random_next = is_random_next
    self.masked_lm_positions = masked_lm_positions
    self.masked_lm_labels = masked_lm_labels

  def __str__(self):
    s = ""
    s += "tokens: %s\n" % (" ".join(
        [tokenization.printable_text(x) for x in self.tokens]))
    s += "segment_ids: %s\n" % (" ".join([str(x) for x in self.segment_ids]))
    s += "is_random_next: %s\n" % self.is_random_next
    s += "masked_lm_positions: %s\n" % (" ".join(
        [str(x) for x in self.masked_lm_positions]))
    s += "masked_lm_labels: %s\n" % (" ".join(
        [tokenization.printable_text(x) for x in self.masked_lm_labels]))
    s += "\n"
    return s

  def __repr__(self):
    return self.__str__()


def pack_i32(obj):
  i64 = np.array(obj, np.int64)
  i32 = i64.astype(np.int32)
  assert (i32 == i64).all()
  return i32.tobytes()


def pack_f32(obj):
  arr = np.array(obj, np.float32)
  return arr.tobytes()


def write_instance_to_example_files(instances, tokenizer, max_seq_length,
                                    max_predictions_per_seq, output_dir,
                                    compression, hashes, size_limit):
  """Create MDS example files from `TrainingInstance`s."""
  total_written = 0

  columns = {
    'input_ids': 'bytes',
    'input_mask': 'bytes',
    'segment_ids': 'bytes',
    'masked_lm_positions': 'bytes',
    'masked_lm_ids': 'bytes',
    'masked_lm_weights': 'bytes',
    'next_sentence_labels': 'bytes',
  }

  writer = MDSWriter(output_dir, columns, compression, hashes, size_limit)

  for (inst_index, instance) in enumerate(instances):
    input_ids = tokenizer.convert_tokens_to_ids(instance.tokens)
    input_mask = [1] * len(input_ids)
    segment_ids = list(instance.segment_ids)
    assert len(input_ids) <= max_seq_length

    while len(input_ids) < max_seq_length:
      input_ids.append(0)
      input_mask.append(0)
      segment_ids.append(0)

    assert len(input_ids) == max_seq_length
    assert len(input_mask) == max_seq_length
    assert len(segment_ids) == max_seq_length

    masked_lm_positions = list(instance.masked_lm_positions)
    masked_lm_ids = tokenizer.convert_tokens_to_ids(instance.masked_lm_labels)
    masked_lm_weights = [1.0] * len(masked_lm_ids)

    while len(masked_lm_positions) < max_predictions_per_seq:
      masked_lm_positions.append(0)
      masked_lm_ids.append(0)
      masked_lm_weights.append(0.0)

    next_sentence_label = 1 if instance.is_random_next else 0

    sample = {
      'input_ids': pack_i32(input_ids),
      'input_mask': pack_i32(input_mask),
      'segment_ids': pack_i32(segment_ids),
      'masked_lm_positions': pack_i32(masked_lm_positions),
      'masked_lm_ids': pack_i32(masked_lm_ids),
      'masked_lm_weights': pack_f32(masked_lm_weights),
      'next_sentence_labels': pack_i32([next_sentence_label]),
    }
    writer.write(sample)
    total_written += 1

    if inst_index < 20:
      tf.logging.info("*** Example ***")
      tf.logging.info("tokens: %s" % " ".join(
          [tokenization.printable_text(x) for x in instance.tokens]))

      for feature_name in sorted(sample):
        feature = sample[feature_name]
        if feature_name == 'masked_lm_weights':
          dtype = np.float32
        else:
          dtype = np.int32
        values = np.frombuffer(feature, dtype)
        tf.logging.info(
            "%s: %s" % (feature_name, " ".join([str(x) for x in values])))

  writer.finish()

  tf.logging.info("Wrote %d total instances", total_written)


def create_int_feature(values):
  feature = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
  return feature


def create_float_feature(values):
  feature = tf.train.Feature(float_list=tf.train.FloatList(value=list(values)))
  return feature


def create_training_instances(input_files, tokenizer, max_seq_length,
                              dupe_factor, short_seq_prob, masked_lm_prob,
                              max_predictions_per_seq, rng):
  """Create `TrainingInstance`s from raw text."""
  all_documents = [[]]

  # Input file format:
  # (1) One sentence per line. These should ideally be actual sentences, not
  # entire paragraphs or arbitrary spans of text. (Because we use the
  # sentence boundaries for the "next sentence prediction" task).
  # (2) Blank lines between documents. Document boundaries are needed so
  # that the "next sentence prediction" task doesn't span between documents.
  for input_file in input_files:
    with tf.gfile.GFile(input_file, "r") as reader:
      while True:
        line = tokenization.convert_to_unicode(reader.readline())
        if not line:
          break
        line = line.strip()

        # Empty lines are used as document delimiters
        if not line:
          all_documents.append([])
        tokens = tokenizer.tokenize(line)
        if tokens:
          all_documents[-1].append(tokens)

  # Remove empty documents
  all_documents = [x for x in all_documents if x]
  rng.shuffle(all_documents)

  vocab_words = list(tokenizer.vocab.keys())
  instances = []
  for _ in range(dupe_factor):
    for document_index in range(len(all_documents)):
      instances.extend(
          create_instances_from_document(
              all_documents, document_index, max_seq_length, short_seq_prob,
              masked_lm_prob, max_predictions_per_seq, vocab_words, rng))

  rng.shuffle(instances)
  return instances


def create_instances_from_document(
    all_documents, document_index, max_seq_length, short_seq_prob,
    masked_lm_prob, max_predictions_per_seq, vocab_words, rng):
  """Creates `TrainingInstance`s for a single document."""
  document = all_documents[document_index]

  # Account for [CLS], [SEP], [SEP]
  max_num_tokens = max_seq_length - 3

  # We *usually* want to fill up the entire sequence since we are padding
  # to `max_seq_length` anyways, so short sequences are generally wasted
  # computation. However, we *sometimes*
  # (i.e., short_seq_prob == 0.1 == 10% of the time) want to use shorter
  # sequences to minimize the mismatch between pre-training and fine-tuning.
  # The `target_seq_length` is just a rough target however, whereas
  # `max_seq_length` is a hard limit.
  target_seq_length = max_num_tokens
  if rng.random() < short_seq_prob:
    target_seq_length = rng.randint(2, max_num_tokens)

  # We DON'T just concatenate all of the tokens from a document into a long
  # sequence and choose an arbitrary split point because this would make the
  # next sentence prediction task too easy. Instead, we split the input into
  # segments "A" and "B" based on the actual "sentences" provided by the user
  # input.
  instances = []
  current_chunk = []
  current_length = 0
  i = 0
  while i < len(document):
    segment = document[i]
    current_chunk.append(segment)
    current_length += len(segment)
    if i == len(document) - 1 or current_length >= target_seq_length:
      if current_chunk:
        # `a_end` is how many segments from `current_chunk` go into the `A`
        # (first) sentence.
        a_end = 1
        if len(current_chunk) >= 2:
          a_end = rng.randint(1, len(current_chunk) - 1)

        tokens_a = []
        for j in range(a_end):
          tokens_a.extend(current_chunk[j])

        tokens_b = []
        # Random next
        is_random_next = False
        if len(current_chunk) == 1 or rng.random() < 0.5:
          is_random_next = True
          target_b_length = target_seq_length - len(tokens_a)

          # This should rarely go for more than one iteration for large
          # corpora. However, just to be careful, we try to make sure that
          # the random document is not the same as the document
          # we're processing.
          for _ in range(10):
            random_document_index = rng.randint(0, len(all_documents) - 1)
            if random_document_index != document_index:
              break

          random_document = all_documents[random_document_index]
          random_start = rng.randint(0, len(random_document) - 1)
          for j in range(random_start, len(random_document)):
            tokens_b.extend(random_document[j])
            if len(tokens_b) >= target_b_length:
              break
          # We didn't actually use these segments so we "put them back" so
          # they don't go to waste.
          num_unused_segments = len(current_chunk) - a_end
          i -= num_unused_segments
        # Actual next
        else:
          is_random_next = False
          for j in range(a_end, len(current_chunk)):
            tokens_b.extend(current_chunk[j])
        truncate_seq_pair(tokens_a, tokens_b, max_num_tokens, rng)

        assert len(tokens_a) >= 1
        assert len(tokens_b) >= 1

        tokens = []
        segment_ids = []
        tokens.append("[CLS]")
        segment_ids.append(0)
        for token in tokens_a:
          tokens.append(token)
          segment_ids.append(0)

        tokens.append("[SEP]")
        segment_ids.append(0)

        for token in tokens_b:
          tokens.append(token)
          segment_ids.append(1)
        tokens.append("[SEP]")
        segment_ids.append(1)

        (tokens, masked_lm_positions,
         masked_lm_labels) = create_masked_lm_predictions(
             tokens, masked_lm_prob, max_predictions_per_seq, vocab_words, rng)
        instance = TrainingInstance(
            tokens=tokens,
            segment_ids=segment_ids,
            is_random_next=is_random_next,
            masked_lm_positions=masked_lm_positions,
            masked_lm_labels=masked_lm_labels)
        instances.append(instance)
      current_chunk = []
      current_length = 0
    i += 1

  return instances


MaskedLmInstance = collections.namedtuple("MaskedLmInstance",
                                          ["index", "label"])


def create_masked_lm_predictions(tokens, masked_lm_prob,
                                 max_predictions_per_seq, vocab_words, rng):
  """Creates the predictions for the masked LM objective."""

  cand_indexes = []
  for (i, token) in enumerate(tokens):
    if token == "[CLS]" or token == "[SEP]":
      continue
    cand_indexes.append(i)

  rng.shuffle(cand_indexes)

  output_tokens = list(tokens)

  num_to_predict = min(max_predictions_per_seq,
                       max(1, int(round(len(tokens) * masked_lm_prob))))

  masked_lms = []
  covered_indexes = set()
  for index in cand_indexes:
    if len(masked_lms) >= num_to_predict:
      break
    if index in covered_indexes:
      continue
    covered_indexes.add(index)

    masked_token = None
    # 80% of the time, replace with [MASK]
    if rng.random() < 0.8:
      masked_token = "[MASK]"
    else:
      # 10% of the time, keep original
      if rng.random() < 0.5:
        masked_token = tokens[index]
      # 10% of the time, replace with random word
      else:
        masked_token = vocab_words[rng.randint(0, len(vocab_words) - 1)]

    output_tokens[index] = masked_token

    masked_lms.append(MaskedLmInstance(index=index, label=tokens[index]))

  masked_lms = sorted(masked_lms, key=lambda x: x.index)

  masked_lm_positions = []
  masked_lm_labels = []
  for p in masked_lms:
    masked_lm_positions.append(p.index)
    masked_lm_labels.append(p.label)

  return (output_tokens, masked_lm_positions, masked_lm_labels)


def truncate_seq_pair(tokens_a, tokens_b, max_num_tokens, rng):
  """Truncates a pair of sequences to a maximum sequence length."""
  while True:
    total_length = len(tokens_a) + len(tokens_b)
    if total_length <= max_num_tokens:
      break

    trunc_tokens = tokens_a if len(tokens_a) > len(tokens_b) else tokens_b
    assert len(trunc_tokens) >= 1

    # We want to sometimes truncate from the front and sometimes from the
    # back to add more randomness and avoid biases.
    if rng.random() < 0.5:
      del trunc_tokens[0]
    else:
      trunc_tokens.pop()


def main(_):
  tf.logging.set_verbosity(tf.logging.INFO)

  tokenizer = tokenization.FullTokenizer(
      vocab_file=FLAGS.vocab_file, do_lower_case=FLAGS.do_lower_case)

  input_files = []
  for input_pattern in FLAGS.input_file.split(","):
    input_files.extend(tf.gfile.Glob(input_pattern))

  tf.logging.info("*** Reading from input files ***")
  for input_file in input_files:
    tf.logging.info("  %s", input_file)

  rng = random.Random(FLAGS.random_seed)
  instances = create_training_instances(
      input_files, tokenizer, FLAGS.max_seq_length, FLAGS.dupe_factor,
      FLAGS.short_seq_prob, FLAGS.masked_lm_prob, FLAGS.max_predictions_per_seq,
      rng)

  tf.logging.info("*** Writing to output directory: %s ***", FLAGS.output_dir)

  hashes = FLAGS.hashes.split(',') if FLAGS.hashes else []
  write_instance_to_example_files(instances, tokenizer, FLAGS.max_seq_length,
                                  FLAGS.max_predictions_per_seq,
                                  FLAGS.output_dir, FLAGS.compression, hashes,
                                  FLAGS.size_limit)


if __name__ == "__main__":
  flags.mark_flag_as_required("input_file")
  flags.mark_flag_as_required("output_dir")
  flags.mark_flag_as_required("vocab_file")
  tf.app.run()



"""Script for picking certain number of samples."""

from argparse import ArgumentParser

from streaming import StreamingDataset, MDSWriter


def parse_args():
    args = ArgumentParser()
    args.add_argument(
        '--in_root',
        type=str,
        required=True,
        help='Local directory path of the input raw dataset',
    )
    args.add_argument(
        '--out_root',
        type=str,
        required=True,
        help='Directory path to store the output dataset',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='zstd:16',
        help='Compression algorithm to use. Default: zstd:16',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh3_64',
        help='Hashing algorithms to apply to shard files. Default: sha1,xxh3_64',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 26,
        help='Shard size limit, after which point to start a new shard. Default: 1 << 26',
    )
    args.add_argument(
        '--num_examples_to_pick',
        type=int,
        default=10000,
        help='Number of examples to select. Default: 10_000',)
    return args.parse_args()


def main(args):
    dataset = StreamingDataset(local=args.in_root, shuffle=False)
    columns = {
        'input_ids': 'bytes',
        'input_mask': 'bytes',
        'segment_ids': 'bytes',
        'masked_lm_positions': 'bytes',
        'masked_lm_ids': 'bytes',
        'masked_lm_weights': 'bytes',
        'next_sentence_labels': 'bytes',
    }
    hashes = args.hashes.split(',') if args.hashes else []
    with MDSWriter(out=args.out_root, 
                   columns=columns, 
                   compression=args.compression, 
                   hashes=hashes, 
                   size_limit=args.size_limit) as writer:
        pick_ratio = dataset.index.total_samples / args.num_examples_to_pick
        for i in range(args.num_examples_to_pick):
            sample = dataset[int(i * pick_ratio)]
            writer.write(sample)


if __name__ ==  '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Post-process a set of subdirectories containing shards into one unified dataset."""

from argparse import ArgumentParser, Namespace
from glob import glob
import json
import os
import shutil


def parse_args() -> Namespace:
    """Parse commmand-line arguments.

    Returns:
        Namespace: Command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument('--in_root', type=str, required=True,
                      help='Location of input shard directories named like group-###')
    args.add_argument('--out_root', type=str, required=True,
                      help='Location of merged shards (a valid streaming dataset)')
    return args.parse_args()


def main(args: Namespace) -> None:
    """Post-process a set of shard subdirectories into one unified dataset.

    Args:
        args (Namespace): Command-line arguments.
    """
    os.makedirs(args.out_root)
    pattern = os.path.join(args.in_root, 'group-*')
    subdirs = sorted(glob(pattern))
    offset = 0
    infos = []
    for subdir in subdirs:
        shards_this_group = len(os.listdir(subdir)) - 1

        # Move shard files.
        for shard in range(shards_this_group):
            old_filename = os.path.join(subdir, f'shard.{shard:05d}.mds.zstd')
            new_filename = os.path.join(args.out_root, f'shard.{offset + shard:05d}.mds.zstd')
            os.rename(old_filename, new_filename)

        # Collect shard infos.
        index_filename = os.path.join(subdir, 'index.json')
        obj = json.load(open(index_filename))
        infos += obj['shards']

        # Update offset.
        offset += shards_this_group

    # Update the indices of the collected shard infos to be global.
    for shard, info in enumerate(infos):
        info['raw_data']['basename'] = f'shard.{shard:05d}.mds'
        info['zip_data']['basename'] = f'shard.{shard:05d}.mds.zstd'

    # Create new index.
    obj = {
        'version': 2,
        'shards': infos,
    }
    index_filename = os.path.join(args.out_root, 'index.json')
    with open(index_filename, 'w') as out:
        json.dump(obj, out)

    # Remove leftover old index files.
    shutil.rmtree(args.in_root)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""MosaicML Streaming Datasets for cloud-native model training."""

from streaming.base.dataloader import StreamingDataLoader
from streaming.base.dataset import StreamingDataset
from streaming.base.format import CSVWriter, JSONWriter, MDSWriter, TSVWriter, XSVWriter
from streaming.base.local import LocalDataset
from streaming.base.stream import Stream

__all__ = [
    'StreamingDataLoader', 'Stream', 'StreamingDataset', 'CSVWriter', 'JSONWriter', 'LocalDataset',
    'MDSWriter', 'TSVWriter', 'XSVWriter'
]



"""Script for picking certain number of sampels.
"""

import argparse
import time
import logging
import collections
import tensorflow.compat.v1 as tf

parser = argparse.ArgumentParser(
    description="Eval sample picker for BERT.")
parser.add_argument(
    '--input_tfrecord',
    type=str,
    default='',
    help='Input tfrecord path')
parser.add_argument(
    '--output_tfrecord',
    type=str,
    default='',
    help='Output tfrecord path')
parser.add_argument(
    '--num_examples_to_pick',
    type=int,
    default=10000,
    help='Number of examples to pick')
parser.add_argument(
    '--max_seq_length',
    type=int,
    default=512,
    help='The maximum number of tokens within a sequence.')
parser.add_argument(
    '--max_predictions_per_seq',
    type=int,
    default=76,
    help='The maximum number of predictions within a sequence.')
args = parser.parse_args()

max_seq_length = args.max_seq_length
max_predictions_per_seq = args.max_predictions_per_seq
logging.basicConfig(level=logging.INFO)

def decode_record(record):
  """Decodes a record to a TensorFlow example."""
  name_to_features = {
      "input_ids":
          tf.FixedLenFeature([max_seq_length], tf.int64),
      "input_mask":
          tf.FixedLenFeature([max_seq_length], tf.int64),
      "segment_ids":
          tf.FixedLenFeature([max_seq_length], tf.int64),
      "masked_lm_positions":
          tf.FixedLenFeature([max_predictions_per_seq], tf.int64),
      "masked_lm_ids":
          tf.FixedLenFeature([max_predictions_per_seq], tf.int64),
      "masked_lm_weights":
          tf.FixedLenFeature([max_predictions_per_seq], tf.float32),
      "next_sentence_labels":
          tf.FixedLenFeature([1], tf.int64),
  }

  example = tf.parse_single_example(record, name_to_features)

  # tf.Example only supports tf.int64, but the TPU only supports tf.int32.
  # So cast all int64 to int32.
  for name in list(example.keys()):
    t = example[name]
    if t.dtype == tf.int64:
      t = tf.to_int32(t)
    example[name] = t

  return example


def create_int_feature(values):
  feature = tf.train.Feature(int64_list=tf.train.Int64List(value=list(values)))
  return feature


def create_float_feature(values):
  feature = tf.train.Feature(float_list=tf.train.FloatList(value=list(values)))
  return feature


if __name__ == '__main__':
  tic = time.time()
  tf.enable_eager_execution()

  d = tf.data.TFRecordDataset(args.input_tfrecord)
  num_examples = 0
  records = []
  for record in d:
    records.append(record)
    num_examples += 1

  writer = tf.python_io.TFRecordWriter(args.output_tfrecord)
  i = 0
  pick_ratio = num_examples / args.num_examples_to_pick
  num_examples_picked = 0
  for i in range(args.num_examples_to_pick):
    example = decode_record(records[int(i * pick_ratio)])
    features = collections.OrderedDict()
    features["input_ids"] = create_int_feature(
        example["input_ids"].numpy())
    features["input_mask"] = create_int_feature(
        example["input_mask"].numpy())
    features["segment_ids"] = create_int_feature(
        example["segment_ids"].numpy())
    features["masked_lm_positions"] = create_int_feature(
        example["masked_lm_positions"].numpy())
    features["masked_lm_ids"] = create_int_feature(
        example["masked_lm_ids"].numpy())
    features["masked_lm_weights"] = create_float_feature(
        example["masked_lm_weights"].numpy())
    features["next_sentence_labels"] = create_int_feature(
        example["next_sentence_labels"].numpy())

    tf_example = tf.train.Example(features=tf.train.Features(feature=features))
    writer.write(tf_example.SerializeToString())
    num_examples_picked += 1

  writer.close()
  toc = time.time()
  logging.info("Picked %d examples out of %d samples in %.2f sec",
               num_examples_picked, num_examples, toc - tic)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Apportion shards/samples to nodes/ranks/workers for elastically deterministic sample order."""

import logging
from typing import Optional

import numpy as np
from numpy.typing import NDArray

from streaming.base.partition.orig import get_partitions_orig
from streaming.base.partition.relaxed import get_partitions_relaxed

logger = logging.getLogger(__name__)

algos = {
    'orig': get_partitions_orig,
    'relaxed': get_partitions_relaxed,
}


def get_partitions(algo: str,
                   num_samples: int,
                   num_canonical_nodes: int,
                   num_physical_nodes: int,
                   ranks_per_node: int,
                   workers_per_rank: int,
                   batch_size: int,
                   drop_first: int = 0,
                   initial_physical_nodes: Optional[int] = None) -> NDArray[np.int64]:
    """Partition the given number of samples to nodes, ranks, and workers.

    Either canonical or physical nodes must be evenly divisible by the other.

    It is suggested to set num_canonical_nodes higher than your expected number of physical nodes,
    because scaling your number of nodes below that level may result in more shards being used
    across node boundaries due to preserving the same global sample order.

    Args:
        algo (str): Partition algorithm name.
        num_samples (int): Dataset size.
        num_canonical_nodes (int): Number of canonical nodes.
        num_physical_nodes (int): Number of physical nodes.
        ranks_per_node (int): Number of ranks per node.
        workers_per_rank (int): Number of worker partitions per rank.
        batch_size (int): Batch size of DataLoader and dataset, which affects how the dataset is
            partitioned over the workers.
        drop_first (int): Number of samples seen already, which are dropped. Defaults to ``0``.
        initial_physical_nodes (int, optional): Number of physical nodes at the start of training.
            Defaults to ``None``.

    Returns:
        NDArray[np.int64]: Partitions of shape (physical nodes, ranks per node, workers per rank,
            batches per worker, batch size).
    """
    world_size = ranks_per_node * num_physical_nodes
    num_repeated_samples = world_size - (num_samples % world_size)
    if num_samples + num_repeated_samples < drop_first:
        raise ValueError(f'Resuming further into the dataset ({drop_first}) than it has samples ' +
                         f'({num_samples})')

    if num_repeated_samples > 0:
        logger.debug(f'Using {num_repeated_samples} repeated samples to ensure that the epoch ' +
                     f'size is divisible by the number of total devices. This ensures that each ' +
                     f'device contributes the same number of samples per global batch. ')

    get = algos[algo]
    return get(num_samples, num_canonical_nodes, num_physical_nodes, ranks_per_node,
               workers_per_rank, batch_size, drop_first, initial_physical_nodes)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Pile streaming dataset conversion script."""

import json
import os
from argparse import ArgumentParser, Namespace
from collections import Counter
from glob import glob
from multiprocessing import Pool
from typing import Iterator

from streaming.base import MDSWriter
from streaming.base.util import get_list_arg


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Args:
        Namespace: command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--in_root',
        type=str,
        required=True,
        help='Local directory path of the input raw dataset',
    )
    args.add_argument(
        '--out_root',
        type=str,
        required=True,
        help='Directory path to store the output dataset',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='zstd:16',
        help='Compression algorithm to use. Empirically, Zstandard has the best performance in ' +
        'our benchmarks. Tune the compression level (from 1 to 22) to trade off time for ' +
        'quality. Defaults to zstd:16',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh64',
        help='Hashing algorithms to apply to shard files. Default: sha1,xxh64',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 27,
        help='Shard size limit, after which point to start a new shard. Default: 1 << 27',
    )
    return args.parse_args()


def each_task(in_root: str, out_root: str, compression: str, hashes: list[str], size_limit: int,
              in_files: list[str]) -> Iterator[tuple[str, str, str, list[str], int]]:
    """Get the arg tuple corresponding to each JSONL input file to convert to streaming.

    Args:
        in_root (str): Root directory of input JSONL files.
        out_root (str): Root directory of output MDS files.
        compression (str): Which compression algorithm to use, or empty if none.
        hashes (List[str]): Hashing algorithms to apply to shard files.
        size_limit (int): Shard size limit, after which point to start a new shard.
        in_files (List[str]): List of input files to generate arguments for.

    Returns:
        Iterator[Tuple[str, str, str, List[str], int]]: Each argument tuple.
    """
    for in_file in in_files:
        assert in_file.startswith(in_root)
        assert in_file.endswith('.jsonl')
        out_dir = os.path.join(out_root, in_file[len(in_root):-len('.jsonl')].lstrip('/'))
        yield in_file, out_dir, compression, hashes, size_limit


def file_to_dir(args: tuple[str, str, str, list[str], int]) -> dict[str, int]:
    """Convert a JSONL input file into a directory of MDS shards.

    This is the unit of work executed by the process pool.

    Args:
        args (Tuple[str, str, str, List[str], int]): All arguments, packed into a tuple because
            process pools only pass one argument.

    Raises:
        ValueError: Invalid sample fields.
        ValueError: Invalid sample meta fields.

    Returns:
        Dict[str, int]: Count of how many samples belonged to each Pile dataset subset.
    """
    in_file, out_dir, compression, hashes, size_limit = args

    columns = {
        'text': 'str',
        'pile_set_name': 'str',
    }

    counts = Counter()
    with MDSWriter(out=out_dir,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit,
                   progress_bar=True) as out:
        for line in open(in_file):
            obj = json.loads(line)
            if sorted(obj.keys()) != ['meta', 'text']:
                raise ValueError('Invalid sample fields.')
            text = obj['text']
            meta = obj['meta']
            if sorted(meta.keys()) != ['pile_set_name']:
                raise ValueError('Invalid sample meta fields.')
            pile_set_name = meta['pile_set_name']
            sample = {
                'text': text,
                'pile_set_name': pile_set_name,
            }
            out.write(sample)
            counts[pile_set_name] += 1
    return counts


def with_id(basename: str, shard_id: int) -> str:
    """Get a new basename with the given shard_id.

    Args:
        basename (str): Old basename of file.
        shard_id (int): New shard ID.

    Returns:
        str: New basename of file.
    """
    parts = basename.split('.')
    parts[1] = f'{shard_id:05}'
    return '.'.join(parts)


def merge_shard_groups(root: str) -> None:
    """Merge ephemeral sub-datasets created in parallel into one dataset.

    Args:
        root (str): Root directory.
    """
    pattern = os.path.join(root, '*')
    subdirs = sorted(glob(pattern))
    shard_id = 0
    infos = []
    for subdir in subdirs:
        index_filename = os.path.join(subdir, 'index.json')
        obj = json.load(open(index_filename))
        for info in obj['shards']:
            old_basename = info['raw_data']['basename']
            new_basename = with_id(old_basename, shard_id)
            info['raw_data']['basename'] = new_basename

            old_basename = info['zip_data']['basename']
            new_basename = with_id(old_basename, shard_id)
            info['zip_data']['basename'] = new_basename

            old_filename = os.path.join(subdir, old_basename)
            new_filename = os.path.join(root, new_basename)
            assert not os.rename(old_filename, new_filename)

            shard_id += 1
            infos.append(info)

        assert not os.remove(index_filename)
        assert not os.rmdir(subdir)

    index_filename = os.path.join(root, 'index.json')
    obj = {
        'version': 2,
        'shards': infos,
    }
    text = json.dumps(obj, sort_keys=True)
    with open(index_filename, 'w') as out:
        out.write(text)


def main(args: Namespace) -> None:
    """Convert the Pile to streaming format.

    Args:
        args (Namespace): Command-line arguments.
    """
    hashes = get_list_arg(args.hashes)

    # Find the original JSONL files to convert.
    pattern = os.path.join(args.in_root, 'train', '*.jsonl')
    trains = sorted(glob(pattern))
    val = os.path.join(args.in_root, 'val.jsonl')
    test = os.path.join(args.in_root, 'test.jsonl')
    in_files = trains + [val, test]

    # Get the arguments for each JSONL file conversion.
    arg_tuples = each_task(args.in_root, args.out_root, args.compression, hashes, args.size_limit,
                           in_files)

    # Process each JSONL file in parallel into directories of shards.
    with Pool() as pool:
        counters = pool.imap(file_to_dir, arg_tuples)
        for in_file, counts in zip(in_files, counters):
            obj = {
                'file': in_file,
                'counts': counts,
            }
            print(json.dumps(obj, sort_keys=True))

    # Merge shard groups.
    train_root = os.path.join(args.out_root, 'train')
    merge_shard_groups(train_root)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Natively supported NLP datasets."""

from streaming.text.c4 import StreamingC4 as StreamingC4
from streaming.text.enwiki import StreamingEnWiki as StreamingEnWiki
from streaming.text.pile import StreamingPile as StreamingPile

__all__ = ['StreamingPile', 'StreamingC4', 'StreamingEnWiki']



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

from tempfile import TemporaryDirectory

import torch
from torch.utils.data import DataLoader

from streaming import MDSWriter
from streaming.base.local import LocalDataset


def test_local_dataset():
    columns = {'value': 'int'}
    num_samples = 100
    with TemporaryDirectory() as dirname:
        with MDSWriter(out=dirname, columns=columns) as out:
            for i in range(num_samples):
                out.write({'value': i})

        dataset = LocalDataset(dirname)
        for sample_id in range(num_samples):
            sample = dataset[sample_id]
            assert sample['value'] == sample_id


def test_local_dataloader():
    columns = {'value': 'int'}
    num_samples = 100
    with TemporaryDirectory() as dirname:
        with MDSWriter(out=dirname, columns=columns) as out:
            for i in range(num_samples):
                out.write({'value': i})

        dataset = LocalDataset(dirname)
        loader = DataLoader(dataset, batch_size=1)
        for sample_id, batch in enumerate(loader):
            assert batch['value'] == torch.tensor([sample_id])



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Convert and upload LAION-400M parquet shards."""

import json
import os
from argparse import ArgumentParser, Namespace
from time import sleep, time
from typing import Iterator, Optional, Union

import numpy as np
from pyarrow import parquet as pq

from streaming import MDSWriter
from streaming.base.storage import CloudUploader


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Returns:
        Namespace: Command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument('--local',
                      type=str,
                      required=True,
                      help='Local directory containing downloaded shards in parquet format.')
    args.add_argument('--remote',
                      type=str,
                      required=True,
                      help='Remote path to upload MDS-formatted shards to.')
    args.add_argument('--keep_parquet',
                      type=int,
                      default=1,
                      help='Whether to keep the parquet shards after conversion (about 10TB).')
    args.add_argument('--keep_mds',
                      type=int,
                      default=1,
                      help='Whether to keep the MDS shards after upload (about 10TB).')
    args.add_argument('--hashes',
                      type=str,
                      default='sha1,xxh64',
                      help='Hashes for validating shards, if any.')
    args.add_argument('--poll_interval',
                      type=float,
                      default=30,
                      help='Interval between polling for newly downloaded shards to process.')
    return args.parse_args()


def each_downloaded_shard(local: str) -> Iterator[int]:
    """Iterate over each downloaded shard.

    Args:
        local (str): Local directory containing shards.

    Returns:
        Iterator[int]: Each downloaded shard ID.
    """
    basenames = set(os.listdir(local))
    count = len(list(filter(lambda s: s.endswith('_stats.json'), basenames)))
    for idx in range(count):
        stats_filename = os.path.join(local, f'{idx:05}_stats.json')
        if not os.path.exists(stats_filename):
            raise RuntimeError('Stats file is missing.')
        yield idx


def get_int(x: Union[float, int]) -> int:
    """Get an int field from pandas.

    Args:
        x (Union[float, int]): The pandas field.

    Returns:
        int: The normalized field.
    """
    if np.isnan(x):
        return 0
    else:
        return int(x)


def get_float(x: float) -> float:
    """Get a float field from pandas.

    Args:
        x (float): The pandas field.

    Returns:
        float: The normalized field.
    """
    return x


def get_bytes(x: Optional[bytes]) -> bytes:
    """Get a bytes field from pandas.

    Args:
        x (bytes, optional): The pandas field.

    Returns:
        float: The normalized field.
    """
    return x or b''


def get_str(x: Optional[str]) -> str:
    """Get a str field from pandas.

    Args:
        x (str, optional): The pandas field.

    Returns:
        str: The normalized field.
    """
    return x or ''


def convert(parquet_filename: str, mds_dirname: str, hashes: list[str]) -> None:
    """Convert a parquet shard to MDS shard with an index.

    Args:
        parquet_filename (str): Filename of the input parquet shard.
        mds_dirname (str): Dirname containing the output MDS shard and index.
        hashes (List[str]): List of hashes to apply to the MDS shard, if any.
    """
    columns = {
        'nsfw': 'str',
        'similarity': 'float64',
        'license': 'str',
        'caption': 'str',
        'url': 'str',
        'key': 'str',
        'status': 'str',
        'error_message': 'str',
        'width': 'int32',
        'height': 'int32',
        'original_width': 'int32',
        'original_height': 'int32',
        'exif': 'str',
        'jpg': 'bytes',
    }
    compression = None  # Don't compress because the vast majority of the data is JPEG.
    size_limit = None  # Put everything in one shard (1:1 mapping of parquet to MDS shards).

    with MDSWriter(out=mds_dirname,
                   columns=columns,
                   compression=compression,
                   hashes=hashes,
                   size_limit=size_limit) as out:
        table = pq.read_table(parquet_filename)
        nr = table.num_rows
        table = table.to_pandas()
        for i in range(nr):
            x = table.iloc[i]
            sample = {
                'nsfw': get_str(x['NSFW']),
                'similarity': get_float(x['similarity']),
                'license': get_str(x['LICENSE']),
                'caption': get_str(x['caption']),
                'url': get_str(x['url']),
                'key': get_str(x['key']),
                'status': get_str(x['status']),
                'error_message': get_str(x['error_message']),
                'width': get_int(x['width']),
                'height': get_int(x['height']),
                'original_width': get_int(x['original_width']),
                'original_height': get_int(x['original_height']),
                'exif': get_str(x['exif']),
                'jpg': get_bytes(x['jpg']),
            }
            out.write(sample)


def upload(local: str, remote: str) -> None:
    """Upload a shard to remote storage.

    Args:
        local (str): Path on local filesystem.
        remote (str): Path on remote filesystem.
    """
    local_dir = os.path.dirname(local)
    remote_dir = os.path.dirname(remote)
    filename = os.path.basename(local)
    cl_uploader = CloudUploader(out=(local_dir, remote_dir))
    cl_uploader.upload_file(filename)


def convert_and_upload_shards(args: Namespace) -> bool:
    """Process any newly downloaded shards.

    Args:
        args (Namespace): Command-line arguments.

    Returns:
        bool: Whether shard downloading is done.
    """
    hashes = args.hashes.split(',') if args.hashes else []
    for idx in each_downloaded_shard(args.local):
        # If the shard is already done, skip it.
        done_filename = os.path.join(args.local, f'{idx:05}.done')
        if os.path.exists(done_filename):
            print(f'Shard {idx:05}: done')
            continue

        # The shard is not marked done, so possibly convert and definitely upload it.
        parquet_filename = os.path.join(args.local, f'{idx:05}.parquet')
        mds_dirname = os.path.join(args.local, f'{idx:05}.mds')
        mds_shard_filename = os.path.join(mds_dirname, 'shard.00000.mds')
        mds_index_filename = os.path.join(mds_dirname, 'index.json')
        remote_shard_filename = os.path.join(args.remote, f'shard.{idx:05}.mds')
        if os.path.exists(mds_dirname):
            if not os.path.exists(mds_shard_filename):
                raise RuntimeError(f'MDS shard file is missing: {mds_shard_filename}.')
            if not os.path.exists(mds_index_filename):
                raise RuntimeError(f'MDS index file is missing: {mds_index_filename}.')
        else:
            print(f'Shard {idx:05}: converting...')
            convert(parquet_filename, mds_dirname, hashes)
        print(f'Shard {idx:05}: uploading...')
        # Rename the file from `shard.00000.mds` to `shard.{idx:05}.mds` since the `upload`
        # method relies on the same filename for source and destination.
        rename_mds_shard_filename = os.path.join(mds_dirname, f'shard.{idx:05}.mds')
        os.rename(mds_shard_filename, rename_mds_shard_filename)
        mds_shard_filename = rename_mds_shard_filename
        upload(mds_shard_filename, remote_shard_filename)
        with open(done_filename, 'w') as out:
            out.write('')

        # Clean up after ourselves to save local storage.
        if not args.keep_parquet:
            os.remove(parquet_filename)
        if not args.keep_mds:
            os.remove(mds_shard_filename)
        print(f'Shard {idx:05}: done')

    # Check for the "done" marker.
    filename = os.path.join(args.local, 'done')
    return os.path.exists(filename)


def collect_and_upload_index(args: Namespace) -> None:
    """Finally, collect and upload the index.

    Args:
        args (Namespace): Command-line arguments.
    """
    infos = []
    for idx in each_downloaded_shard(args.local):
        sub_index_filename = os.path.join(args.local, f'{idx:05}.mds', 'index.json')
        obj = json.load(open(sub_index_filename))
        info, = obj['shards']
        infos.append(info)

    obj = {
        'version': 2,
        'shards': infos,
    }
    local = os.path.join(args.local, 'index.json')
    with open(local, 'w') as out:
        json.dump(obj, out)

    remote = os.path.join(args.remote, 'index.json')
    upload(local, remote)


def main(args: Namespace) -> None:
    """Convert and upload shards as they are created.

    Args:
        args (Namespace): Command-line arguments.
    """
    while True:
        last_poll = time()
        is_done = convert_and_upload_shards(args)
        if is_done:
            break
        now = time()
        elapsed = now - last_poll
        if elapsed < args.poll_interval:
            sleep(args.poll_interval - elapsed)

    collect_and_upload_index(args)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import json
import os
import tempfile
import time
import urllib.parse
from multiprocessing.shared_memory import SharedMemory as BuiltinSharedMemory
from typing import Optional, Union

import pytest

from streaming.base.constant import RESUME
from streaming.base.shared.prefix import _get_path
from streaming.base.storage.download import download_file
from streaming.base.storage.upload import CloudUploader
from streaming.base.util import (bytes_to_int, clean_stale_shared_memory, get_list_arg,
                                 merge_index, number_abbrev_to_int, retry)

MY_PREFIX = 'train_' + str(time.time())
MY_BUCKET = {
    'gs://': 'testing-bucket',
    's3://': 'testing-bucket',
    'oci://': 'testing-bucket',
}
os.environ[
    'OBJC_DISABLE_INITIALIZE_FORK_SAFETY'] = 'YES'  # set to yes to all fork process in spark calls


@pytest.mark.parametrize(('text', 'expected_output'), [('hello,world', ['hello', 'world']),
                                                       ('hello', ['hello']), ('', [])])
def test_get_list_arg(text: str, expected_output: list[Optional[str]]):
    output = get_list_arg(text)
    assert output == expected_output


def test_bytes_to_int():
    input_to_expected = [
        ('1234', 1234),
        ('1b', 1),
        ('50b', 50),
        ('50B', 50),
        ('100kb', 102400),
        (' 100 kb', 102400),
        ('75mb', 78643200),
        ('75MB', 78643200),
        ('75 mb ', 78643200),
        ('1.39gb', 1492501135),
        ('1.39Gb', 1492501135),
        ('2tb', 2199023255552),
        ('3pb', 3377699720527872),
        ('1.11eb', 1279742870113600256),
        ('1.09zb', 1286844866581978415104),
        ('2.0yb', 2417851639229258349412352),
        (1234, 1234),
        (1, 1),
        (0.5 * 1024, 512),
        (100 * 1024, 102400),
        (75 * 1024**2, 78643200),
        (75 * 1024 * 1024, 78643200),
        (35.78, 35),
        (325388903.203984, 325388903),
    ]
    for size_pair in input_to_expected:
        output = bytes_to_int(size_pair[0])
        assert output == size_pair[1]


def test_bytes_to_int_Exception():
    input_data = ['', '12kbb', '27mxb', '79kkb']
    for value in input_data:
        with pytest.raises(ValueError, match=f'Unsupported value/suffix.*'):
            _ = bytes_to_int(value)


def test_number_abbrev_to_int():
    input_to_expected = [
        ('1234', 1234),
        ('1k', 1000),
        ('50k', 50000),
        ('50K', 50000),
        ('100k', 100000),
        (' 100 k', 100000),
        ('75m', 75000000),
        ('75M', 75000000),
        ('75 m ', 75000000),
        ('1.39b', 1390000000),
        ('1.39B', 1390000000),
        ('2t', 2000000000000),
        ('3 T', 3000000000000),
        (1234, 1234),
        (1, 1),
        (0.5 * 1000, 500),
        (100 * 1000, 100000),
        (75 * 1000**2, 75000000),
        (75 * 1000 * 1000, 75000000),
        (35.78, 35),
        (325388903.203984, 325388903),
    ]
    for size_pair in input_to_expected:
        output = number_abbrev_to_int(size_pair[0])
        assert output == size_pair[1]


def test_number_abbrev_to_int_Exception():
    input_data = ['', '12kbb', '27mxb', '79bk', '79bb', '79 b    m', 'p 64', '64p']
    for value in input_data:
        with pytest.raises(ValueError, match=f'Unsupported value/suffix.*'):
            _ = number_abbrev_to_int(value)


def test_clean_stale_shared_memory():
    # Clean up the stale shared memory
    clean_stale_shared_memory()

    # Create a leaked shared memory
    name = _get_path(0, RESUME)
    _ = BuiltinSharedMemory(name, True, 64)

    # Clean up the stale shared memory
    clean_stale_shared_memory()

    # If clean up is successful, it should raise FileNotFoundError Exception
    with pytest.raises(FileNotFoundError):
        _ = BuiltinSharedMemory(name, False, 64)


def integrity_check(out: Union[str, tuple[str, str]],
                    keep_local: bool,
                    expected_n_shard_files: int = -1):
    """Check if merged_index file has integrity
        If merged_index is a cloud url, first download it to a temp local file.

    Args:
        out (Union[str, Tuple[str,str]]): folder that merged index.json resides
        keep_local: whether to check local file
        expected_n_shard_files (int): If -1, find the number in `out` with get_expected()
    """

    def get_expected(mds_root: str):
        n_shard_files = 0
        cu = CloudUploader.get(mds_root, exist_ok=True, keep_local=True)
        for o in cu.list_objects():
            # ends with .mds, or perhaps .mds.zstd
            if '.mds' in o:
                n_shard_files += 1
        return n_shard_files

    cu = CloudUploader.get(out, keep_local=True, exist_ok=True)

    with tempfile.TemporaryDirectory() as temp_dir:
        if cu.remote:
            download_file(os.path.join(cu.remote, 'index.json'),
                          os.path.join(temp_dir, 'index.json'),
                          timeout=60)
            if expected_n_shard_files == -1:
                expected_n_shard_files = get_expected(cu.remote)
            local_merged_index_path = os.path.join(temp_dir, 'index.json')
        else:
            local_merged_index_path = os.path.join(cu.local, 'index.json')
            if expected_n_shard_files == -1:
                expected_n_shard_files = get_expected(cu.local)

        if not keep_local:
            assert not os.path.exists(os.path.join(cu.local, 'index.json'))
            return

        assert os.path.exists(
            local_merged_index_path
        ), f'{local_merged_index_path} does not exist when keep_local is {keep_local}'
        merged_index = json.load(open(local_merged_index_path, 'r'))
        n_shard_files = len({b['raw_data']['basename'] for b in merged_index['shards']})
        assert n_shard_files == expected_n_shard_files, f'expected {expected_n_shard_files} shard files but got {n_shard_files}'


@pytest.mark.parametrize('scheme', ['gs', 's3', 'oci', 'dbfs'])
def test_format_remote_index_files(scheme: str):
    """Validate the format of remote index files."""
    from streaming.base.util import _format_remote_index_files

    if scheme == 'dbfs':
        remote = os.path.join('dbfs:/', 'Volumes')
    else:
        full_scheme = scheme + '://'
        remote = os.path.join(full_scheme, MY_BUCKET[full_scheme], MY_PREFIX)

    index_files = [
        os.path.join('folder_1', 'index.json'),
        os.path.join('folder_2', 'index.json'),
        os.path.join('folder_3', 'index.json'),
    ]
    remote_index_files = _format_remote_index_files(remote, index_files)

    assert len(remote_index_files) == len(index_files)
    for file in remote_index_files:
        obj = urllib.parse.urlparse(file)
        assert obj.scheme == scheme


@pytest.mark.parametrize('index_file_urls_pattern', [1, 2, 3])
@pytest.mark.parametrize('keep_local', [True, False])
@pytest.mark.parametrize('scheme', ['gs://', 's3://', 'oci://'])
def test_merge_index_from_list_local(local_remote_dir: tuple[str, str], keep_local: bool,
                                     index_file_urls_pattern: int, scheme: str):
    """Validate the final merge index json for following patterns of index_file_urls:
    1. All URLs are str (local). All URLs are accessible locally -> no download
    2. All URLs are str (local). At least one url is unaccessible locally -> Error
    3. All URLs are tuple (local, remote). All URLs are accessible locally -> no download
    4. All URLs are tuple (local, remote). At least one url is not accessible locally -> download all
    5. All URLs are str (remote) -> download all
    """
    from decimal import Decimal

    from pyspark.sql import SparkSession
    from pyspark.sql.types import DecimalType, IntegerType, StringType, StructField, StructType

    from streaming.base.converters import dataframeToMDS

    def not_merged_index(index_file_path: str, out: str):
        """Check if index_file_path is the merged index at folder out."""
        prefix = str(urllib.parse.urlparse(out).path)
        return os.path.dirname(index_file_path).strip('/') != prefix.strip('/')

    local, _ = local_remote_dir

    mds_out = out = local

    spark = SparkSession.builder.getOrCreate()  # pyright: ignore
    schema = StructType([
        StructField('id', IntegerType(), nullable=False),
        StructField('name', StringType(), nullable=False),
        StructField('amount', DecimalType(10, 2), nullable=False)
    ])
    data = [(1, 'Alice', Decimal('123.45')), (2, 'Bob', Decimal('67.89')),
            (3, 'Charlie', Decimal('987.65'))]
    df = spark.createDataFrame(data=data, schema=schema).repartition(3)
    mds_kwargs = {'out': mds_out, 'columns': {'id': 'int32', 'name': 'str'}, 'keep_local': True}
    dataframeToMDS(df, merge_index=False, mds_kwargs=mds_kwargs)

    local_cu = CloudUploader.get(local, exist_ok=True, keep_local=True)
    local_index_files = [
        o for o in local_cu.list_objects() if o.endswith('.json') and not_merged_index(o, local)
    ]

    if index_file_urls_pattern == 1:
        merge_index(local_index_files, out, keep_local=keep_local)

    if index_file_urls_pattern == 2:
        with tempfile.TemporaryDirectory() as a_temporary_folder:
            index_file_urls = [
                os.path.join(a_temporary_folder, os.path.basename(s)) for s in local_index_files
            ]
            with pytest.raises(RuntimeError, match=f'.*Failed to download index.json.*'):
                merge_index(index_file_urls, out, keep_local=keep_local)

        with tempfile.TemporaryDirectory() as a_temporary_folder:
            index_file_urls = [(os.path.join(a_temporary_folder, os.path.basename(s)), '')
                               for s in local_index_files]
            with pytest.raises(FileNotFoundError, match=f'.*Check data availability!.*'):
                merge_index(index_file_urls, out, keep_local=keep_local)
            return

    if index_file_urls_pattern == 3:
        remote_index_files = [
            os.path.join(scheme, MY_BUCKET[scheme], MY_PREFIX, os.path.basename(o))
            for o in local_index_files
            if o.endswith('.json') and not_merged_index(o, local)
        ]
        index_file_urls = list(zip(local_index_files, remote_index_files))
        merge_index(index_file_urls, out, keep_local=keep_local)

    integrity_check(out, keep_local=keep_local)


@pytest.mark.parametrize('n_partitions', [1, 2, 3, 4])
@pytest.mark.parametrize('keep_local', [False, True])
def test_merge_index_from_root_local(local_remote_dir: tuple[str, str], n_partitions: int,
                                     keep_local: bool):
    from decimal import Decimal

    from pyspark.sql import SparkSession
    from pyspark.sql.types import DecimalType, IntegerType, StringType, StructField, StructType

    from streaming.base.converters import dataframeToMDS

    out, _ = local_remote_dir

    spark = SparkSession.builder.getOrCreate()  # pyright: ignore
    schema = StructType([
        StructField('id', IntegerType(), nullable=False),
        StructField('name', StringType(), nullable=False),
        StructField('amount', DecimalType(10, 2), nullable=False)
    ])

    data = [(1, 'Alice', Decimal('123.45')), (2, 'Bob', Decimal('67.89')),
            (3, 'Charlie', Decimal('987.65'))]

    df = spark.createDataFrame(data=data, schema=schema).repartition(n_partitions)

    mds_kwargs = {'out': out, 'columns': {'id': 'int32', 'name': 'str'}, 'keep_local': keep_local}

    mds_path, _ = dataframeToMDS(df, merge_index=False, mds_kwargs=mds_kwargs)
    merge_index(mds_path, keep_local=keep_local)
    integrity_check(mds_path, keep_local=keep_local)


@pytest.mark.parametrize('with_args', [True, False])
def test_retry(with_args: bool):
    num_tries = 0
    return_after = 2

    if with_args:
        decorator = retry(RuntimeError, num_attempts=3, initial_backoff=0.01, max_jitter=0.01)
        return_after = 2
    else:
        decorator = retry
        # Need to return immediately to avoid timeouts
        return_after = 0

    @decorator
    def flaky_function():
        nonlocal num_tries
        if num_tries < return_after:
            num_tries += 1
            raise RuntimeError('Called too soon!')
        return "Third time's a charm"

    assert flaky_function() == "Third time's a charm"



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""COCO 2017 streaming dataset conversion scripts."""

import json
import os
from argparse import ArgumentParser, Namespace
from typing import Iterable

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset
from tqdm import tqdm

from streaming.base import MDSWriter
from streaming.base.util import get_list_arg


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Args:
        Namespace: command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--in_root',
        type=str,
        required=True,
        help='Local directory path of the input raw dataset',
    )
    args.add_argument(
        '--out_root',
        type=str,
        required=True,
        help='Directory path to store the output dataset',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='',
        help='Compression algorithm to use. Default: None',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh64',
        help='Hashing algorithms to apply to shard files. Default: sha1,xxh64',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 25,
        help='Shard size limit, after which point to start a new shard. Default: 1 << 25',
    )
    args.add_argument(
        '--progress_bar',
        type=int,
        default=1,
        help='tqdm progress bar. Default: 1 (Act as True)',
    )
    args.add_argument(
        '--leave',
        type=int,
        default=0,
        help='Keeps all traces of the progressbar upon termination of iteration. Default: 0 ' +
        '(Act as False)',
    )
    return args.parse_args()


class _COCODetection(Dataset):
    """PyTorch Dataset for the COCO dataset.

    Args:
        img_folder (str): Path to a COCO folder.
        annotate_file (str): Path to a file that contains image id, annotations (e.g., bounding
            boxes and object classes) etc.
    """

    def __init__(self, img_folder: str, annotate_file: str):
        self.img_folder = img_folder
        self.annotate_file = annotate_file

        # Start processing annotation
        with open(annotate_file) as fin:
            self.data = json.load(fin)

        self.images = {}

        self.label_map = {}
        self.label_info = {}
        # 0 stands for the background
        cnt = 0
        self.label_info[cnt] = 'background'
        for cat in self.data['categories']:
            cnt += 1
            self.label_map[cat['id']] = cnt
            self.label_info[cnt] = cat['name']

        # build inference for images
        for img in self.data['images']:
            img_id = img['id']
            img_name = img['file_name']
            img_size = (img['height'], img['width'])
            if img_id in self.images:
                raise Exception('duplicated image record')
            self.images[img_id] = (img_name, img_size, [])

        # read bboxes
        for bboxes in self.data['annotations']:
            img_id = bboxes['image_id']
            bbox = bboxes['bbox']
            bbox_label = self.label_map[bboxes['category_id']]
            self.images[img_id][2].append((bbox, bbox_label))

        for k, v in list(self.images.items()):
            if len(v[2]) == 0:
                self.images.pop(k)

        self.img_keys = list(self.images.keys())

    def labelnum(self):
        return len(self.label_info)

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx: int):
        img_id = self.img_keys[idx]
        img_data = self.images[img_id]
        fn = img_data[0]
        img_path = os.path.join(self.img_folder, fn)

        img = Image.open(img_path).convert('RGB')

        htot, wtot = img_data[1]
        bbox_sizes = []
        bbox_labels = []

        for (l, t, w, h), bbox_label in img_data[2]:
            r = l + w
            b = t + h
            bbox_size = (l / wtot, t / htot, r / wtot, b / htot)
            bbox_sizes.append(bbox_size)
            bbox_labels.append(bbox_label)

        bbox_sizes = torch.tensor(bbox_sizes)
        bbox_labels = torch.tensor(bbox_labels)

        return img, img_id, (htot, wtot), bbox_sizes, bbox_labels


def each(dataset: _COCODetection, shuffle: bool) -> Iterable[dict[str, bytes]]:
    """Generator over each dataset sample.

    Args:
        dataset (_COCODetection): COCO detection dataset.
        shuffle (bool): Whether to shuffle the samples.

    Yields:
        Iterator[Iterable[Dict[str, bytes]]]: Sample dicts.
    """
    if shuffle:
        indices = np.random.permutation(len(dataset))
    else:
        indices = np.arange(len(dataset))
    for idx in indices:
        _, img_id, (htot, wtot), bbox_sizes, bbox_labels = dataset[idx]

        img_id = dataset.img_keys[idx]
        img_data = dataset.images[img_id]
        img_basename = img_data[0]
        img_filename = os.path.join(dataset.img_folder, img_basename)
        img_bytes = open(img_filename, 'rb').read()

        yield {
            'img': img_bytes,
            'img_id': img_id,
            'htot': htot,
            'wtot': wtot,
            'bbox_sizes': np.array(bbox_sizes, dtype=np.float32),  # (_,4) np.float32 array.
            'bbox_labels': np.array(bbox_labels, dtype=np.int64),  # np.int64 array.
        }


def main(args: Namespace) -> None:
    """Main: create COCO streaming dataset.

    Args:
        args (Namespace): command-line arguments.

    Raises:
        FileNotFoundError: Images path does not exist.
        FileNotFoundError: Annotations file does not exist.
        ValueError: Number of samples in a dataset does not match.
    """
    columns = {
        'img': 'jpeg',
        'img_id': 'int',
        'htot': 'int',
        'wtot': 'int',
        'bbox_sizes': 'ndarray:float32',
        'bbox_labels': 'ndarray:int64',
    }

    for (split, expected_num_samples, shuffle) in [
        ('train', 117266, True),
        ('val', 4952, False),
    ]:
        out_split_dir = os.path.join(args.out_root, split)

        split_images_in_dir = os.path.join(args.in_root, f'{split}2017')
        if not os.path.exists(split_images_in_dir):
            raise FileNotFoundError(f'Images path does not exist: {split_images_in_dir}')
        split_annotations_in_file = os.path.join(args.in_root, 'annotations',
                                                 f'instances_{split}2017.json')
        if not os.path.exists(split_annotations_in_file):
            raise FileNotFoundError(
                f'Annotations file does not exist: {split_annotations_in_file}')
        dataset = _COCODetection(split_images_in_dir, split_annotations_in_file)

        if len(dataset) != expected_num_samples:
            raise ValueError(f'Number of samples in a dataset doesn\'t match. Expected ' +
                             f'{expected_num_samples}, but got {len(dataset)}')

        hashes = get_list_arg(args.hashes)

        if args.progress_bar:
            dataset = tqdm(each(dataset, shuffle), leave=args.leave, total=len(dataset))
        else:
            dataset = each(dataset, shuffle)

        with MDSWriter(out=out_split_dir,
                       columns=columns,
                       compression=args.compression,
                       hashes=hashes,
                       size_limit=args.size_limit,
                       progress_bar=args.progress_bar) as out:
            for sample in dataset:
                out.write(sample)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""An ordered set that can be used as an LRU cache."""

from collections import OrderedDict
from typing import Any


class LastUsedOrderedSet(OrderedDict):
    """An ordered dict that can be used as an LRU cache.

    This is a subclass of OrderedDict, with some LRU-specific functions and all values as ``None``.
    """

    def setitem(self, key: Any, move_to_end: bool = True):
        """Set/add an item.

        Args:
            key (Any): key to be added.
            move_to_end (bool, optional): whether to move the item to the end, signifying most
                recent access. Defaults to ``True``.
        """
        super().__setitem__(key, None)
        self.move_to_end(key, last=move_to_end)

    def popLRU(self):
        """Pop the least recently used item (located at the front)."""
        return self.popitem(last=False)[0]

    def setuse(self, key: Any):
        """Mark an item as used, moving it to the end.

        Args:
            key (Any): key of element to move to the end, signifying most recent access.
        """
        self.setitem(key)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""C4 (Colossal Cleaned Common Crawl) streaming dataset conversion script."""

import os
from argparse import ArgumentParser, Namespace
from typing import Any, Iterator

import datasets
from datasets.arrow_dataset import Dataset
from torch.utils.data import DataLoader, IterableDataset, get_worker_info
from tqdm import tqdm

from streaming.base import MDSWriter
from streaming.base.util import get_list_arg


def parse_args() -> Namespace:
    """Parse command-line arguments.

    Args:
        Namespace: command-line arguments.
    """
    args = ArgumentParser()
    args.add_argument(
        '--out_root',
        type=str,
        required=True,
        help='Directory path to store the output dataset',
    )
    args.add_argument(
        '--compression',
        type=str,
        default='zstd:7',
        help='Compression algorithm to use. Default: zstd:7',
    )
    args.add_argument(
        '--hashes',
        type=str,
        default='sha1,xxh64',
        help='Hashing algorithms to apply to shard files. Default: sha1,xxh64',
    )
    args.add_argument(
        '--size_limit',
        type=int,
        default=1 << 27,
        help='Shard size limit, after which point to start a new shard. Default: 1 << 27',
    )
    args.add_argument(
        '--batch_size',
        type=int,
        default=512,
        help='DataLoader batch size. Default: 512',
    )
    args.add_argument(
        '--progress_bar',
        type=int,
        default=1,
        help='tqdm progress bar. Default: 1 (True)',
    )
    args.add_argument(
        '--leave',
        type=int,
        default=0,
        help='Keeps all traces of the progressbar upon termination of iteration. Default: 0 ' +
        '(False)',
    )
    return args.parse_args()


def get(split: str) -> IterableDataset:
    """Collect the samples for this dataset split.

    Args:
        split (str): Dataset split name.

    Raises:
        ValueError: Shards must divide evenly by num_workers

    Returns:
        IterableDataset: An IterableDataset.
    """

    class ShardedC4(IterableDataset):

        def __init__(self):
            self.dataset = datasets.load_dataset(  # pyright: ignore
                path='c4', name='en', split=split, streaming=True)

        def num_shards(self):
            return len(self.dataset._ex_iterable.kwargs['filepaths'])  # pyright: ignore

        def __iter__(self):
            worker_info = get_worker_info()
            if worker_info:
                num_workers = worker_info.num_workers
                worker_id = worker_info.id
                shards = self.dataset._ex_iterable.kwargs['filepaths']  # pyright: ignore
                if len(shards) % num_workers:
                    raise ValueError('Shards must divide evenly by num_workers')
                self.dataset._ex_iterable.kwargs['filepaths'] = (  # pyright: ignore
                    shards[worker_id::num_workers])
            return iter(self.dataset)

    return ShardedC4()


def each(dataset: Dataset, num_workers: int, batch_size: int) -> Iterator[dict[str, Any]]:
    """Iterate over each dataset sample.

    Args:
        dataset (Dataset): A HuggingFace Dataset locally downloaded.
        num_workers (int): DataLoader number of workers.
        batch_size (int): DataLoader batch size.

    Returns:
        Iterator[Dict[str, Any]]: Sample dicts.
    """
    prefetch_factor = max(1, 2 * batch_size // num_workers)
    loader = DataLoader(
        dataset=dataset,  # pyright: ignore
        batch_size=batch_size,
        num_workers=num_workers,
        prefetch_factor=prefetch_factor)
    for batch in loader:
        keys = list(batch.keys())
        current_bs = len(batch[keys[0]])
        for idx in range(current_bs):
            yield {key: batch_values[idx] for key, batch_values in batch.items()}


def main(args: Namespace) -> None:
    """Main: create streaming C4 dataset.

    Args:
        args (Namespace): command-line arguments.
    """
    splits = [
        ('train', 'train', 364868892, 64),
        ('validation', 'val', 364608, 8),
    ]
    columns = {'text': 'str', 'timestamp': 'str', 'url': 'str'}
    hashes = get_list_arg(args.hashes)
    for old_split, new_split, num_samples, num_workers in splits:
        dataset = get(old_split)
        split_dir = os.path.join(args.out_root, new_split)
        with MDSWriter(out=split_dir,
                       columns=columns,
                       compression=args.compression,
                       hashes=hashes,
                       size_limit=args.size_limit,
                       progress_bar=args.progress_bar) as out:
            samples = each(dataset, num_workers, args.batch_size)  # pyright: ignore
            if args.progress_bar:
                samples = tqdm(samples, total=num_samples, leave=args.leave)
            for sample in samples:
                out.write(sample)


if __name__ == '__main__':
    main(parse_args())



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Apportion shards/samples such that device batches have samples only from a single stream."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray

from streaming.base.partition import get_partitions
from streaming.base.shuffle import get_shuffle
from streaming.base.world import World

if TYPE_CHECKING:
    from streaming.base.dataset import StreamingDataset

logger = logging.getLogger(__name__)


def generate_work_device_per_stream_batching(dataset: StreamingDataset, world: World, epoch: int,
                                             sample_in_epoch: int) -> NDArray[np.int64]:
    """Generate this epoch's arrangement of samples for ``device_per_stream`` batching.

    This is only called in local rank zero. When ``batching_method`` is set to ``device_per_stream``,
    each device batch consists of samples from only one stream.

    Args:
        dataset (StreamingDataset): Dataset to generate the partition for.
        world (World): World state.
        epoch (int): Which epoch it is.
        sample_in_epoch (int): Where we are in the epoch.

    Returns:
        NDArray[np.int64]: The epoch (num physical nodes, ranks per node, workers per rank,
            batches per worker, batch size).
    """
    # Ensure that num_canonical_nodes has been set.
    if dataset.num_canonical_nodes is None:
        raise RuntimeError(f'`num_canonical_nodes` can never be None. ' +
                           f'Provide a positive integer.')

    if dataset.num_canonical_nodes % world.num_nodes != 0:
        raise ValueError(
            f'For `device_per_stream` batching, num_canonical_nodes must be divisible by physical nodes. '
            +
            f'Got {dataset.num_canonical_nodes} canonical nodes and {world.num_nodes} physical nodes.'
        )

    # First, for each stream, sample each shard of the stream according to
    # stream proportions/repeats/samples. We obtain the resampled size of each shard
    # in the stream and a mapping from the training "big" sample ID to the underlying
    # shard "small" sample ID. Then, we also partition each stream's samples over
    # nodes/devices/workers. We handle sample_in_epoch (for resumption) at the end.
    partition_per_stream = []

    batch_size = dataset.batch_size
    assert isinstance(batch_size, int), f'Batch size must be an integer. Got {type(batch_size)}.'

    for stream_id, stream in enumerate(dataset.streams):
        shuffle_units, small_per_big = dataset.resample_streams(epoch, stream_id)
        samples_in_stream = len(small_per_big)
        # The partition for each stream is constructed with batch size 1 and physical nodes
        # equal to canonical nodes in order to make sure that the sample order from each batch
        # stays the same even when the number of physical nodes & batch size change.
        # We later reshape these partitions to the correct batch size per stream.
        # We also handle used samples (drop_first) at the end.
        stream_partition = get_partitions(dataset.partition_algo, samples_in_stream,
                                          dataset.num_canonical_nodes, dataset.num_canonical_nodes,
                                          world.ranks_per_node, world.workers_per_rank, 1, 0,
                                          dataset.initial_physical_nodes)
        if dataset.shuffle:
            # Ratio of stream's shuffle block size to overall shuffle block size should be the
            # same as the ratio of the stream's samples to overall samples.
            # This ensures that the overall training shuffle block size is still approximately
            # equal to what is set by the user, and allows for reasoning about cache_limit as well.
            if not isinstance(dataset.shuffle_block_size, int):
                raise TypeError(f'Dataset `shuffle_block_size` must be an integer. ' +
                                f'Got {type(dataset.shuffle_block_size)} instead.')
            shuffle_block_portion = int(dataset.shuffle_block_size * stream.proportion)
            stream_shuffle = get_shuffle(dataset.shuffle_algo, shuffle_units,
                                         dataset.num_canonical_nodes, dataset.shuffle_seed, epoch,
                                         shuffle_block_portion)
            stream_partition = np.where(stream_partition != -1, stream_shuffle[stream_partition],
                                        -1)
        # The small_per_big array already corresponds to indices of samples per shard of each
        # stream. So each sample ID in the stream's partition already corresponds to the sample ID
        # in the right shard.
        partition_per_stream.append(
            np.where(stream_partition != -1, small_per_big[stream_partition], -1))

    # We now merge the partitions from each stream to get a partition over all streams
    # but for each physical node, where each device batch has samples only from a single stream.
    # Original partitions are (physical nodes, ranks, workers, batches per worker, batch size).
    # Since we have partitioned each stream with physical nodes equal to canonical nodes,
    # we index into the partition to get each stream's portion of samples for each physical node.
    batches_per_stream = []
    batches_from_partitions = []
    ncn_per_node = dataset.num_canonical_nodes // world.num_nodes
    for node in range(world.num_nodes):
        # We keep track of per-node, per-stream device batches, and the number of batches per-node, per-stream.
        per_node_stream_partitions = []
        per_node_batches_per_stream = []
        for stream_idx, partition in enumerate(partition_per_stream):
            # Reshape the partition to be device batches in order of traversal.
            # We only count only batches without -1 in them.
            stream_samples_inorder = partition[node * ncn_per_node:(node + 1) *
                                               ncn_per_node].transpose(3, 2, 0, 1, 4).flatten()
            # Pad samples to make sure they are divisible by the device batch size.
            padding_samples = batch_size - (stream_samples_inorder.size % batch_size)
            stream_samples_inorder = np.concatenate(
                (stream_samples_inorder, np.full(padding_samples, -1)))
            # Reshape samples to be device batches in order of traversal.
            stream_samples_inorder = stream_samples_inorder.reshape(-1, batch_size)
            num_full_batches = np.count_nonzero(np.min(stream_samples_inorder, axis=1) >= 0)
            per_node_batches_per_stream.append(num_full_batches)
            if num_full_batches != stream_samples_inorder.shape[0]:
                logger.warning(
                    f'Because of the `device_per_stream` batching method, some batches with an inadequate '
                    + f'number of samples from stream with index {stream_idx} will be dropped.')
            if num_full_batches > 0:
                per_node_stream_partitions.append(stream_samples_inorder[:num_full_batches])
            else:
                raise ValueError(
                    f'Stream with index {stream_idx} does not have an adequate number of ' +
                    f'samples to construct even a single device batch of size {batch_size}. ' +
                    f'Training will occur without any samples from this stream!')

        batches_per_stream.append(per_node_batches_per_stream)
        batches_from_partitions.append(per_node_stream_partitions)

    # Combine all device batches from all streams into one array, per node.
    all_partition_batches = []
    for node in range(world.num_nodes):
        all_partition_batches.append(np.concatenate(batches_from_partitions[node]))

    # Find the maximum number of device batches per node, for padding purposes.
    max_device_batches_per_node = max(
        [node_batches.shape[0] for node_batches in all_partition_batches])
    # If the maximum number of device batches per node is not divisible by the number of devices, increase
    # it so that it is. This is to ensure that later, we can reshape the device batches to global batches.
    num_devices = world.num_nodes * world.ranks_per_node
    padding_max_device_batches = num_devices - (max_device_batches_per_node % num_devices)
    max_device_batches_per_node += padding_max_device_batches

    # When shuffling is enabled, each epoch has a different rng.
    # Otherwise, we keep the rng the same to keep the sample ordering deterministic.
    epoch_seed = dataset.shuffle_seed + epoch if dataset.epoch_seed_change else dataset.shuffle_seed
    epoch_rng = np.random.default_rng(epoch_seed)

    # Shuffle the device batch origin order for each node.
    for node in range(world.num_nodes):
        # stream_origins is an array that tells us which stream each device batch is using.
        stream_origins = np.concatenate(
            [np.full(n_batch, i) for i, n_batch in enumerate(batches_per_stream[node])])
        epoch_rng.shuffle(stream_origins)

        # Now, we want the batch_indices array to correctly index into the node's device batches
        # array according to stream_origins in order to get our final device batch order.
        # For each stream, we want to traverse its device batches in the same order as its original partition.
        batch_indices = np.zeros(stream_origins.shape[0]).astype(np.int64)
        batch_offset = 0
        for i, n_device_batch in enumerate(batches_per_stream[node]):
            # Update batch_indices for the one stream at a time.
            batch_indices[stream_origins == i] += batch_offset + np.arange(n_device_batch)
            batch_offset += n_device_batch

        # Rearrange the node's device batches array by the batch_indices we have obtained.
        # This is a (num_device_batches, device_batch_size) array.
        all_partition_batches[node] = all_partition_batches[node][batch_indices]

        # If needed, pad the node's device batches array to have the same number of device batches
        # as the maximum number of device batches across all nodes.
        padding_batches = max_device_batches_per_node - all_partition_batches[node].shape[0]
        all_partition_batches[node] = np.concatenate(
            (all_partition_batches[node], np.full((padding_batches, batch_size), -1)))

    # Concatenate all the per-node device batches into one large sample partition across all nodes.
    # This sample partition is all device batches that are seen in the same order as in training.
    # So for example, we go from node 0: [a1, a2, a3], node 1: [b1, b2, b3], node 2: [c1, c2, c3]
    # to [a1, b1, c1, a2, b2, c2, a3, b3, c3], where each entry is a device batch.
    all_partition_batches = np.stack(all_partition_batches, axis=1).reshape(-1, batch_size)

    # If applicable we resume right after the most recently used full global batch.
    global_batch_size = batch_size * world.num_nodes * world.ranks_per_node
    if sample_in_epoch % global_batch_size != 0:
        logger.warning(
            'Because of the `device_per_stream` batching method, resumption may only occur on a sample \
            that is a multiple of the current global batch size of ' + str(global_batch_size) +
            '. Resuming training after the most recently finished global batch.')

    # Pad and reshape all_partition_batches to the global batch size instead of device batch size.
    # padding_batches = (global_batch_size // batch_size) - (all_partition_batches.shape[0] %
    #                                                     (global_batch_size // batch_size))
    # all_partition_batches = np.concatenate(
    #     (all_partition_batches, np.full((padding_batches, batch_size), -1)))
    # Padding to global_batch_size *should* be taken care of by the padding we did earlier, for node device batches.
    all_partition_batches = all_partition_batches.reshape(-1, global_batch_size)

    # Discard previous batches that may have already finished
    resumption_batch = sample_in_epoch // global_batch_size
    all_partition_batches = all_partition_batches[resumption_batch:]

    # Add padding batches if needed to ensure that we have an even number of
    # batches per worker/rank/node.
    current_samples = all_partition_batches.size
    divisibility_requirement = world.num_nodes * world.ranks_per_node * \
          world.workers_per_rank * batch_size
    if current_samples % divisibility_requirement != 0:
        samples_needed = divisibility_requirement - (current_samples % divisibility_requirement)
        padding_batches_needed = samples_needed // global_batch_size
        all_partition_batches = np.concatenate(
            (all_partition_batches, np.full((padding_batches_needed, global_batch_size), -1)))

    # Reverse the transposition and reshape from earlier.
    # Final result is (physical nodes, ranks, worker, batches per worker, batch size).
    return all_partition_batches.reshape(-1, world.workers_per_rank, world.num_nodes,
                                         world.ranks_per_node,
                                         batch_size).transpose(2, 3, 1, 0, 4)



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Module to write and read the dataset in MDS format."""

from streaming.base.format.mds.reader import MDSReader
from streaming.base.format.mds.writer import MDSWriter

__all__ = ['MDSReader', 'MDSWriter']



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Check whether sample encoding is of supported JSON types."""

from abc import ABC, abstractmethod
from typing import Any

__all__ = ['is_json_encoded', 'is_json_encoding']


class Encoding(ABC):
    """Encoding of an object of JSON type."""

    @classmethod
    @abstractmethod
    def is_encoded(cls, obj: Any) -> bool:
        """Get whether the given object is of this type.

        Args:
            obj (Any): Encoded object.

        Returns:
            bool: Whether of this type.
        """
        raise NotImplementedError

    @staticmethod
    def _validate(data: Any, expected_type: Any) -> bool:
        if not isinstance(data, expected_type):
            raise AttributeError(
                f'data should be of type {expected_type}, but instead, found as {type(data)}')
        return True


class Str(Encoding):
    """Store str."""

    @classmethod
    def is_encoded(cls, obj: Any) -> bool:
        return cls._validate(obj, str)


class Int(Encoding):
    """Store int."""

    @classmethod
    def is_encoded(cls, obj: Any) -> bool:
        return cls._validate(obj, int)


class Float(Encoding):
    """Store float."""

    @classmethod
    def is_encoded(cls, obj: Any) -> bool:
        return cls._validate(obj, float)


_encodings = {'str': Str, 'int': Int, 'float': Float}


def is_json_encoded(encoding: str, value: Any) -> bool:
    """Get whether the given object is of this encoding type.

    Args:
        encoding (str): The encoding.
        value (Any): The object.

    Returns:
        bool: Whether of this type.
    """
    cls = _encodings[encoding]
    return cls.is_encoded(value)


def is_json_encoding(encoding: str) -> bool:
    """Get whether the given encoding is supported.

    Args:
        encoding (str): Encoding.

    Returns:
        bool: Whether encoding is supported.
    """
    return encoding in _encodings



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest

from streaming.base.sampling import get_sampling


@pytest.mark.parametrize('use_epoch', [True, False])
def test_choose_per_shard_adds_up(use_epoch: bool):
    for granularity in range(1, 100):
        for _ in range(10):
            samples_per_shard = 100 + np.random.choice(100, 10)
            samples = sum(samples_per_shard)
            choose = np.random.choice(samples)
            seed = np.random.choice(31337)
            epoch = np.random.choice(42)
            choose_per_shard = get_sampling(samples_per_shard, choose, granularity, seed, epoch,
                                            use_epoch)
            assert (0 <= choose_per_shard).all()
            assert (choose_per_shard <= samples_per_shard).all()
            assert sum(choose_per_shard) == choose


@pytest.mark.parametrize('use_epoch', [True, False])
def test_is_deterministic(use_epoch: bool):
    for granularity in range(1, 100):
        for _ in range(3):
            samples_per_shard = 100 + np.random.choice(100, 10)
            samples = sum(samples_per_shard)
            choose = np.random.choice(samples)
            seed = np.random.choice(31337)
            epoch = np.random.choice(42)
            last = None
            for _ in range(2):
                choose_per_shard = get_sampling(samples_per_shard, choose, granularity, seed,
                                                epoch, use_epoch)
                if last is not None:
                    assert (last == choose_per_shard).all()
                last = choose_per_shard


@pytest.mark.parametrize('use_epoch', [True, False])
def test_balance(use_epoch: bool):
    samples_per_shard = 1_000 + np.random.choice(1_000, 10)
    samples = sum(samples_per_shard)
    choose = np.random.choice(samples)
    choose_per_shard = np.zeros(len(samples_per_shard))
    for granularity in range(1, 100):
        for _ in range(10):
            seed = np.random.choice(31337)
            epoch = np.random.choice(42)
            choose_per_shard += get_sampling(samples_per_shard, choose, granularity, seed, epoch,
                                             use_epoch)
    choose_per_shard /= 99 * 10
    rates = choose_per_shard / samples_per_shard
    imbalance = rates.std() / rates.mean()
    assert imbalance < 0.05



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Any, Optional

import pytest

import streaming.base.hashing as shash
from streaming.base import StreamingDataset
from tests.common.utils import convert_to_mds

logger = logging.getLogger(__name__)


def test_get_hashes():
    supported_hashes = [
        'xxh32', 'sha384', 'blake2s', 'md5', 'sha3_384', 'xxh128', 'sha1', 'blake2b', 'xxh3_128',
        'sha256', 'sha224', 'sha512', 'sha3_512', 'xxh3_64', 'xxh64', 'sha3_224', 'sha3_256'
    ]
    hashes = shash.get_hashes()
    assert isinstance(hashes, set)
    assert len(hashes) == len(supported_hashes)
    assert hashes == set(supported_hashes)


@pytest.mark.parametrize(('algo_name', 'is_supported'), [('sha384', True), ('xxh128', True),
                                                         ('', False), ('fake', False)])
def test_is_hash(algo_name: str, is_supported: bool):
    output = shash.is_hash(algo_name)
    assert output is is_supported


@pytest.mark.parametrize(('algo_name', 'data', 'expected_output'), [
    ('md5', b'hello', '5d41402abc4b2a76b9719d911017c592'),
    ('sha3_256', b'hello', '3338be694f50c5f338814986cdf0686453a888b84f424d792af4b9202398f392'),
    ('xxh3_64', b'hello', '9555e8555c62dcfd'),
])
def test_get_hash(algo_name: str, data: bytes, expected_output: str):
    output = shash.get_hash(algo_name, data)
    assert isinstance(output, str)
    assert output == expected_output


@pytest.mark.parametrize(('algo_name', 'data'), [
    ('', b'hello'),
    ('sha3', b'hello'),
])
def test_get_hash_invalid_algo(algo_name: str, data: bytes):
    with pytest.raises(ValueError):
        _ = shash.get_hash(algo_name, data)


@pytest.mark.parametrize('hashes', [None, ['sha1'], ['sha384', 'xxh128']])
@pytest.mark.parametrize('compression', [None, 'zstd:7'])
@pytest.mark.usefixtures('local_remote_dir')
def test_validate_checksum(local_remote_dir: Any, hashes: Optional[list[str]], compression: str):
    num_samples = 117
    remote_dir, local_dir = local_remote_dir
    convert_to_mds(out_root=remote_dir,
                   dataset_name='sequencedataset',
                   num_samples=num_samples,
                   compression=compression,
                   hashes=hashes,
                   size_limit=1 << 8)

    # Build StreamingDataset
    hash = hashes[0] if hashes is not None else None
    dataset = StreamingDataset(local=local_dir,
                               remote=remote_dir,
                               validate_hash=hash,
                               batch_size=1)

    # Append sample ID
    sample_order = []
    for sample in dataset:
        sample_order.append(sample['id'])

    # Test length
    assert len(
        dataset
    ) == num_samples, f'Got dataset length={len(dataset)} samples, expected {num_samples}'



# Copyright 2022-2024 MosaicML Streaming authors
# SPDX-License-Identifier: Apache-2.0

"""Base module for dataset reader and writer."""

from streaming.base.format.base.reader import FileInfo, Reader

__all__ = ['FileInfo', 'Reader']
